from numpy import vectorize
import pandas as pd
import pytest
from sklearn import metrics


from holisticai.datasets import load_dataset

metrics_dict = {
    "Accuracy": metrics.accuracy_score,
    "Balanced accuracy": metrics.balanced_accuracy_score,
    "Precision": metrics.precision_score,
    "Recall": metrics.recall_score,
    "F1-Score": metrics.f1_score,
}

SHARD_SIZE=60

@pytest.fixture
def categorical_dataset():
    dataset = load_dataset("adult", protected_attribute='race') # x,y,group_a,group_b
    dataset = dataset.groupby(['y','group_a']).sample(SHARD_SIZE, random_state=42) # 0-ga | 0-gb  | 1-ga | 1-gb
    return dataset.train_test_split(test_size=0.2, stratify=dataset['y'], random_state=0)

@pytest.fixture
def regression_dataset():
    dataset = load_dataset("us_crime", protected_attribute='race')
    dataset = dataset.groupby('group_a').sample(SHARD_SIZE, random_state=42)
    return dataset.train_test_split(test_size=0.2, random_state=0)

@pytest.fixture
def multiclass_dataset():
    dataset = load_dataset("us_crime_multiclass", protected_attribute='race')
    dataset = dataset.groupby(['y','group_a']).sample(SHARD_SIZE, random_state=42)
    dataset = dataset.map(lambda sample: {'stratify': str(sample['y']) + str(sample['group_a'])}, vectorized=False)
    return dataset.train_test_split(test_size=0.2, stratify=dataset['stratify'], random_state=0)

@pytest.fixture
def recommender_dataset():
    dataset = load_dataset("lastfm")
    dataset = dataset.sample(SHARD_SIZE, random_state=42)
    return dataset.train_test_split(test_size=0.2, stratify=dataset['p_attr'], random_state=0)

@pytest.fixture
def clustering_dataset():
    dataset = load_dataset("clinical_records", protected_attribute='sex')
    return dataset.train_test_split(test_size=0.2, stratify=dataset['y'], random_state=0)

def check_results(df1, df2, atol=1e-5):
    """
    Checks if the results are equal

    Parameters
    ----------
    df1 : pandas.DataFrame
        The first dataframe
    df2 : pandas.DataFrame
        The second dataframe

    Returns
    -------
    pandas.DataFrame
        The dataframe containing the results
    """
    import numpy as np
    assert np.isclose(df1["Value"].iloc[0], df2["Value"].iloc[0], atol=atol), (df1["Value"].iloc[0], df2["Value"].iloc[0])



from __future__ import annotations

import logging
from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np
from holisticai.utils.transformers.bias import BMPreprocessing
from jax.nn import softmax
from scipy.optimize import minimize

logger = logging.getLogger(__name__)


def get_x_hat_y_hat(prototypes, w, x):
    m = softmax(-jnp.linalg.norm(x[:, None] - prototypes, axis=2), axis=1)
    x_hat = jnp.dot(m, prototypes)
    y_hat = jnp.clip(jnp.dot(m, w.reshape((-1, 1))), jnp.finfo(float).eps, 1.0)
    return m, x_hat, y_hat


class ObjectiveFunction:
    def __init__(self, features_dim, verbose, x, y, m_a, m_b, k=10, A_x=0.01, A_y=0.1, A_z=0.5):
        """
        Objective function for optimization

        Parameters
        ----------
        features_dim : int
            Number of features
        verbose : int
            If zero, then no output
        x : array-like
            Input data
        y : array-like
            Target vector
        m_a : array-like
            Mask vector
        m_b : array-like
            Mask vector
        k : int, optional
            Number of prototypes. Default is 10
        A_x : float, optional
            Input reconstruction quality term weight. Default is 0.01
        A_y : float, optional
            Output prediction error. Default is 0.1
        A_z : float, optional
            Fairness constraint term weight. Default is 0.5
        """
        self.verbose = verbose
        self.x = x
        self.y = y
        self.one_minus_y = 1 - y
        self.m_a = m_a
        self.m_b = m_b
        self.k = k
        self.A = jnp.array([A_x, A_y, A_z])
        self.prototypes_shape = (self.k, features_dim)

    def __call__(self, parameters):
        w = parameters[: self.k]
        prototypes = parameters[self.k :].reshape(self.prototypes_shape)

        m, x_hat, y_hat = get_x_hat_y_hat(prototypes, w, self.x)

        dx2 = (x_hat - self.x) ** 2
        loss_x = jnp.mean(dx2[self.m_b]) + jnp.mean(dx2[self.m_a])

        loss_z = jnp.mean(jnp.abs(jnp.mean(m[self.m_b], axis=0) - jnp.mean(m[self.m_a], axis=0)))

        loss_y = -jnp.mean(self.y * jnp.log(y_hat) + self.one_minus_y * jnp.log(1.0 - y_hat))

        loss = jnp.array([loss_x, loss_y, loss_z])

        total_loss = jnp.dot(self.A, loss)

        return total_loss


class LearningFairRepresentation(BMPreprocessing):
    """
    Learning fair representations [1]_ finds a latent representation which encodes the data well\
    while obfuscates information about protected attributes [1].

    Parameters
    ----------
    k : int, optional
        Number of prototypes. Default is 5
    Ax : float, optional
        Input recontruction quality term weight. Default is 0.01
    Ay : float, optional
        Output prediction error. Default is 1.0
    Az : float, optional
        Fairness constraint term weight. Default is 50.0
    maxiter : int, optional
        Maximum number of iterations. Default is 5000
    maxfun : int, optional
        Maximum number of function evaluations. Default is 5000
    verbose : int, optional
        If zero, then no output. Default is 0
    seed : int, optional
        Seed to make `predict` repeatable. Default is None

    References
    ----------
    .. [1] Zemel, Rich, et al. "Learning fair representations."
        International conference on machine learning. PMLR, 2013.
    """

    def __init__(
        self,
        k: Optional[int] = 5,
        Ax: Optional[float] = 0.01,
        Ay: Optional[float] = 1.0,
        Az: Optional[float] = 50.0,
        maxiter: Optional[int] = 5000,
        maxfun: Optional[int] = 5000,
        verbose: Optional[int] = 0,
        seed: Optional[int] = None,
    ):
        self.seed = seed
        self.k = k
        self.Ax = Ax
        self.Ay = Ay
        self.Az = Az

        self.w = None
        self.prototypes = None
        self.learned_model = None
        self.maxiter = maxiter
        self.maxfun = maxfun
        self.problem_type = "binary"
        self.verbose = verbose

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit data to learn a fair representation transform.

        Parameters
        ----------
        X : matrix-like
            Input data
        y : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        ------
            Self
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        params = self._load_data(
            X=X,
            y=y,
            group_a=group_a,
            group_b=group_b,
        )
        y = params["y"].reshape([-1, 1])
        group_a = params["group_a"]
        group_b = params["group_b"]
        x = params["X"]

        features_dim = np.shape(x)[1]

        parameters_initialization = np.random.uniform(size=self.k + features_dim * self.k)
        parameters_bounds = [(0, 1)] * self.k + [(None, None)] * features_dim * self.k
        args = (
            x,
            y,
            group_a == 1,
            group_b == 1,
            self.k,
            self.Ax,
            self.Ay,
            self.Az,
        )

        obj_fun = ObjectiveFunction(features_dim, self.verbose, *args)

        @jax.jit
        def objective(params):
            return obj_fun(params)

        result = minimize(
            objective,
            parameters_initialization,
            method="L-BFGS-B",
            bounds=parameters_bounds,
            options={"maxiter": self.maxiter, "disp": 0},
        )
        self.learned_model = result.x
        self.w = self.learned_model[: self.k]
        self.prototypes = self.learned_model[self.k :].reshape((self.k, features_dim))
        return self

    def transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Transform data to a fair representation

        Parameters
        ----------
        X : matrix-like
            Input data
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        ------
        array-like
            Transformed data
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        x = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        _, x_hat_b, _ = get_x_hat_y_hat(self.prototypes, self.w, x[group_b == 1])
        _, x_hat_a, _ = get_x_hat_y_hat(self.prototypes, self.w, x[group_a == 1])

        new_x = x.copy()
        new_x[group_b == 1] = x_hat_b
        new_x[group_a == 1] = x_hat_a

        return new_x

    def fit_transform(
        self,
        X: np.ndarray,
        y: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit and transform

        Parameters
        ----------
        X : matrix-like
            Input data
        y : array-like
            Target vector
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        ------
            Self
        """
        return self.fit(X, y, group_a, group_b).transform(X, group_a, group_b)



import numpy as np
import pandas as pd
import pytest
from holisticai.explainability.metrics.global_importance._xai_ease_score import XAIEaseScore, XAIEaseAnnotator, compare_tangents
from collections import namedtuple

@pytest.fixture
def xai_ease_annotator():
    return XAIEaseAnnotator()

@pytest.fixture
def xai_ease_score():
    return XAIEaseScore()

def test_compute_xai_ease_score_data(xai_ease_annotator):
    from holisticai.utils import PartialDependence
    partial_dependence = PartialDependence( values = [[
        {'average': [[0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]]},
        {'average': [[0.4, 0.5, 0.6, 0.6, 0.5, 0.4, 0.4, 0.5, 0.6]]},
    ]])
    ranked_feature_importance = namedtuple("ranked_feature_importance", "feature_names")
    ranked_feature_importance.feature_names = ['feature1', 'feature2']
    score_data = xai_ease_annotator.compute_xai_ease_score_data(partial_dependence.values[0], ranked_feature_importance)
    assert len(score_data) == 2
    assert 'feature' in score_data.columns
    assert 'scores' in score_data.columns

def test_compute_xai_ease_score(xai_ease_score):
    score_data = pd.DataFrame({'feature': ['feature1', 'feature2'], 'scores': ['Easy', 'Medium']})
    assert xai_ease_score.compute_xai_ease_score(score_data) == 0.75

def test_xai_feature_ease_score(xai_ease_score):
    from holisticai.utils import PartialDependence
    partial_dependence = PartialDependence( values = [[
        {'average': [[0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]]},
        {'average': [[0.4, 0.5, 0.6, 0.6, 0.5, 0.4, 0.4, 0.5, 0.6]]},
    ]])
    ranked_feature_importance = namedtuple("ranked_feature_importance", "feature_names")
    ranked_feature_importance.feature_names = ['feature1', 'feature2']
    assert xai_ease_score(partial_dependence, ranked_feature_importance) == 0.5

def test_xai_ease_score(xai_ease_score):
    from holisticai.utils import PartialDependence, Importances
    partial_dependence = [[
        {'average': [[0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]], 'grid_values': [[1,2,3,4,5,6,7,8,9]]},
        {'average': [[0.4, 0.5, 0.6, 0.6, 0.5, 0.4, 0.4, 0.5, 0.6]], 'grid_values': [[1,2,3,4,5,6,7,8,9]]},
    ]]
    partial_dependence = PartialDependence(values=partial_dependence)
    feature_importance = Importances(feature_names= ['feature1', 'feature2'], values=np.array([0.5, 0.5]))
    assert xai_ease_score(partial_dependence, feature_importance) == 0.5
    

def test_compare_tangents():
    points = [0.1, 0.2, 0.3, 0.8, 0.10, 0.2, 0.5, 0.4]
    slopes, similarities = compare_tangents(points)
    assert np.isclose(slopes[0], -1.0)
    assert np.isclose(slopes[1], -1.0)


def test_compare_tangents_equal():
    points = [0.1, 0.2, 0.3, 0.1, 0.2, 0.3, 0.1, 0.2, 0.3]
    slopes, similarities = compare_tangents(points)
    assert np.isclose(slopes[0], 1.0)
    assert np.isclose(slopes[1], 1.0)
 


# Imports
import numpy as np
import pandas as pd
from numpy.testing import assert_approx_equal

# Recommender
from holisticai.bias.metrics import (
    aggregate_diversity,
    avg_f1_ratio,
    avg_precision_ratio,
    avg_recall_ratio,
    avg_recommendation_popularity,
    exposure_entropy,
    exposure_kl,
    exposure_l1,
    gini_index,
    mad_score,
    recommender_mae_ratio,
    recommender_rmse_ratio,
)

# Formatting
from holisticai.utils import extract_columns
from tests.bias.utils import load_bias_recommender_data

# Dataset
df_rec = load_bias_recommender_data()

# Format data into numpy
group_a, group_b = extract_columns(df_rec, cols=["group_a", "group_b"])
mat_pred = df_rec[["item_1", "item_2", "item_3", "item_4"]].to_numpy()
mat_true = df_rec[
    ["item_1_true", "item_2_true", "item_3_true", "item_4_true"]
].to_numpy()


def test_aggregate_diversity():
    """test aggregate_diversity"""
    a = aggregate_diversity(mat_pred, top=None, thresh=0.5)
    b = 1
    assert_approx_equal(a, b)


def test_gini_index():
    """test gini_index"""
    a = gini_index(mat_pred, top=None, thresh=0.5, normalize=True)
    b = 0.1851851851851852
    assert_approx_equal(a, b)


def test_exposure_entropy():
    """test exposure_entropy"""
    a = exposure_entropy(mat_pred, top=None, thresh=0.5, normalize=True)
    b = 1.9546859469463558 * np.log(2)
    assert_approx_equal(a, b)


def test_avg_recommendation_popularity():
    """test avg_recommendation_popularity"""
    a = avg_recommendation_popularity(mat_pred, top=None, thresh=0.5, normalize=True)
    b = 4.833333333333334
    assert_approx_equal(a, b)

    mat2 = np.concatenate((mat_pred, np.ones(mat_pred.shape) * 0.1), axis=0)
    a2 = avg_recommendation_popularity(mat2, top=None, thresh=0.5, normalize=True)
    assert_approx_equal(a2, b)


def test_mad_score():
    """test mad_score"""
    a = mad_score(group_a, group_b, mat_pred, normalize=True)
    b = 0.00925925925925919
    assert_approx_equal(a, b)


def test_exposure_l1():
    """test exposure_l1"""
    a = exposure_l1(group_a, group_b, mat_pred)
    b = 0.5 * (
        abs(2 / 7 - 3 / 12)
        + abs(2 / 7 - 4 / 12)
        + abs(1 / 7 - 2 / 12)
        + abs(2 / 7 - 3 / 12)
    )
    assert_approx_equal(a, b)


def test_exposure_kl():
    """test exposure_kl"""
    a = exposure_kl(group_a, group_a, mat_pred)
    b = 0
    assert_approx_equal(a, b)


def test_avg_precision_ratio():
    """test avg_precision_ratio"""
    group_a = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    group_b = np.array([0, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    a = avg_precision_ratio(group_a, group_b, mat_pred, mat_true)
    b = (1 + 2 / 3) / 2
    assert_approx_equal(a, b)


def test_avg_recall_ratio():
    """test avg_recall_ratio"""
    group_a = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    group_b = np.array([0, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    a = avg_recall_ratio(group_a, group_b, mat_pred, mat_true)
    b = 1
    assert_approx_equal(a, b)


def test_avg_f1_ratio():
    """test avg_f1_ratio"""
    group_a = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    group_b = np.array([0, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    a = avg_f1_ratio(group_a, group_b, mat_pred, mat_true)
    b = (2 * (2 / 3) / (2 / 3 + 1) + 1) / 2
    assert_approx_equal(a, b)


def test_recommender_rmse_ratio():
    """test recommender_rmse_ratio"""
    group_a = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    group_b = np.array([0, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    a = recommender_rmse_ratio(group_a, group_b, mat_pred, mat_true)
    b = np.sqrt(0.2**2 + 0.2**2 + 0.5**2) / np.sqrt(0.2**2)
    assert_approx_equal(a, b)


def test_recommender_mae_ratio():
    """test avg_mae_ratio"""
    group_a = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
    group_b = np.array([0, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    a = recommender_mae_ratio(group_a, group_b, mat_pred, mat_true)
    b = 4.499999999999998
    assert_approx_equal(a, b)



from __future__ import annotations

import numpy as np
from holisticai.utils import ConditionalImportances, Importances


class PositionParity:
    name: str = "Position Parity"
    reference: float = 1.0

    def __call__(self, conditional_feature_importance: ConditionalImportances, feature_importance: Importances):
        conditional_position_parity = {}
        for group_name, cond_features in conditional_feature_importance:
            match_order = [c == r for c, r in zip(cond_features.feature_names, feature_importance.feature_names)]
            m_order_cum = np.cumsum(match_order) / np.arange(1, len(match_order) + 1)
            conditional_position_parity[group_name] = np.mean(m_order_cum)
        return float(np.mean(np.mean(list(conditional_position_parity.values()))))


def position_parity(conditional_feature_importance: ConditionalImportances, ranked_feature_importance: Importances):
    """
    This metric, ranging from 0 to 1, measures how well the top feature importances (>80%) maintain their ranking when considering conditional importance for classes (classification) or quantiles (regression).

    Parameters
    ----------
    conditional_feature_importance: ConditionalImportances
        The feature importance for each output label (classification) or output region (regression).
    ranked_feature_importance: Importances
        The ranked feature importance values.

    Returns
    -------
    float
        The position parity metric value.

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.explainability.commons import (
    ...     ConditionalImportances,
    ...     Importances,
    ... )
    >>> from holisticai.explainability.metrics import position_parity
    >>> conditional_feature_importance = ConditionalImportances(
    ...     values={
    ...         "group1": Importances(
    ...             values=np.array([0.40, 0.35, 0.25]),
    ...             feature_names=["feature_1", "feature_2", "feature_3"],
    ...         ),
    ...         "group2": Importances(
    ...             values=np.array([0.50, 0.30, 0.20]),
    ...             feature_names=["feature_3", "feature_2", "feature_1"],
    ...         ),
    ...     }
    ... )
    >>> ranked_feature_importance = Importances(
    ...     values=np.array([0.50, 0.40, 0.10]),
    ...     feature_names=["feature_1", "feature_2", "feature_3"],
    ... )
    >>> position_parity(conditional_feature_importance, ranked_feature_importance)
    0.6388888888888888
    """
    metric = PositionParity()
    return metric(conditional_feature_importance, ranked_feature_importance)



# base imports
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

# utils
from holisticai.utils import get_colors

# range metrics
RANGE_METRICS_REGRESSION = {
    "RMSE Ratio": 0.2,
    "RMSE Ratio Q80": 0.2,
    "MAE Ratio": 0.2,
    "MAE Ratio Q80": 0.2,
    "Correlation Difference": 0.05,
    "Disparate Impact Q90": 0.05,
    "Disparate Impact Q80": 0.05,
    "Disparate Impact Q50": 0.05,
    "Statistical Parity Q50": 0.05,
    "Average Score Ratio": 0.05,
    "Average Score Difference": 0.1,
    "Z Score Difference": 0.1,
    "Max Statistical Parity": 0.1,
    "Statistical Parity AUC": 0.1,
}

RANGE_METRICS_CLASSIFICATION = {
    "Statistical Parity": 0.05,
    "Disparate Impact": 0.05,
    "Four Fifths Rule": 0.05,
    "Cohen D": 0.05,
    "2SD Rule": 0.05,
    "Equality of Opportunity Difference": 0.05,
    "False Positive Rate Difference": 0.05,
    "Average Odds Difference": 0.05,
    "Accuracy Difference": 0.05,
}

RANGE_METRICS_CLUSTERING = {
    "Cluster Balance": 0.05,
    "Minimum Cluster Ratio": 0.05,
    "Cluster Distribution Total Variation": 0.05,
    "Cluster Distribution KL Div": 0.05,
    "Social Fairness Ratio": 0.05,
    "Silhouette Difference": 0.05,
}


RANGE_METRICS = {
    "binary_classification": RANGE_METRICS_CLASSIFICATION,
    "regression": RANGE_METRICS_REGRESSION,
    "clustering": RANGE_METRICS_CLUSTERING,
}


def bias_metrics_report(
    model_type: str,
    table_metrics: pd.DataFrame,
    table_metrics_mitigated: pd.DataFrame = None,
):
    """
    Plot bias report for different model types.

    Parameters
    ----------
    model_type : str
        Type of model: 'binary_classification', 'regression', 'clustering'
    table_metrics : pandas.DataFrame
        Dataframe containing bias metrics.
    table_metrics_mitigated : bool, optional
        Whether the bias metrics are for mitigated model or not, by default False
    """
    metric_names = list(table_metrics.index)

    if table_metrics_mitigated is None:
        metrics_biased = table_metrics.copy()
        columns = ["Baseline", "Reference"]
        columns_plot = ["Baseline"]
        metrics_biased.columns = columns
        fill_range = [-1, 1]

    else:
        metrics_biased = pd.concat(
            [table_metrics["Value"], table_metrics_mitigated[["Value", "Reference"]]],
            axis=1,
        )
        columns = ["Baseline", "Mitigator", "Reference"]
        columns_plot = ["Baseline", "Mitigator"]
        metrics_biased.columns = columns
        fill_range = [-0.5, 1.5]

    cols = 4
    rows = len(metric_names) // cols
    if len(metric_names) % cols != 0:
        rows += 1

    max_cols_per_rows = 3
    fig_size = (12, 7) if rows >= max_cols_per_rows else (12, 4)

    sns.set_style("darkgrid")
    fig, axes = plt.subplots(ncols=cols, nrows=rows, figsize=fig_size)

    for i, name in enumerate(metric_names):
        metric_data = metrics_biased[metrics_biased.index == name]
        row, col = divmod(i, cols)
        sns.barplot(
            data=metric_data[columns_plot],
            palette=get_colors(2),
            ax=axes[row, col],
        )
        axes[row, col].set_title(name)
        axes[row, col].axhline(y=metric_data["Reference"].values[0], color="black", linestyle="--")
        axes[row, col].set_ylabel("Score")

        if i == len(metric_names) - 1 and i % cols != cols - 1:
            for j in range(i % cols + 1, cols):
                axes[row, j].axis("off")

        elif i == len(metric_names) - 1 and i % cols == cols - 1:
            for j in range(cols):
                axes[row + 1, j].axis("off")

        if name != "No Disparate Impact Level":
            axes[row, col].fill_between(
                fill_range,
                metric_data["Reference"].values[0] - RANGE_METRICS[model_type][name],
                metric_data["Reference"].values[0] + RANGE_METRICS[model_type][name],
                color="slategray",
                alpha=0.4,
            )

        if i == 0:
            axes[row, col].legend(
                [
                    plt.Line2D([0], [0], linestyle="--", color="black", lw=2, label="Reference"),
                    plt.Rectangle(
                        [0, 0],
                        1,
                        1,
                        fc="slategray",
                        alpha=0.4,
                        ec="None",
                        label="Range",
                    ),
                ],
                ["Reference", "Fair Area"],
                loc="upper left",
                bbox_to_anchor=(-1, 1.0),
            )

        plt.tight_layout()



from __future__ import annotations

from typing import Any, Callable, Union

import pandas as pd
from holisticai.security.commons import BlackBoxAttack
from holisticai.security.metrics._utils import check_valid_output_type
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.metrics import accuracy_score, f1_score, mean_absolute_error, mean_squared_error


def to_numerical_or_categorical(y: pd.Series):
    if y.dtype.kind in ["f"]:
        return y

    if y.dtype.kind in ["i", "u"]:
        return y.astype("category")

    if len(y.unique()) < 2:
        raise ValueError("The target variable must have more than 1 unique value")
    return y.astype("category")


class AttributeAttackScore:
    reference: float = 0
    name: str = "Attribute Attack Score"

    def __call__(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
        attribute_attack: str,
        attack_train_ratio: float = 0.5,
        metric_fn: Union[str, Callable, None] = None,
        attacker_estimator: Any = None,
    ) -> float:
        check_valid_output_type(y_train)

        y_train = to_numerical_or_categorical(y_train)
        y_test = to_numerical_or_categorical(y_test)

        if attacker_estimator is None:
            has_continous_values = x_train[attribute_attack].dtype.kind in ["i", "u", "f"]
            attacker_estimator = LinearRegression() if has_continous_values else LogisticRegression()
            if metric_fn is None:
                metric_fn = mean_squared_error if has_continous_values else accuracy_score

        if isinstance(metric_fn, str):
            if metric_fn == "accuracy":
                metric_fn = accuracy_score
            if metric_fn == "f1":
                metric_fn = f1_score
            if metric_fn == "mean_squared_error":
                metric_fn = mean_squared_error
            if metric_fn == "mean_absolute_error":
                metric_fn = mean_absolute_error

        attacker = BlackBoxAttack(
            attacker_estimator=attacker_estimator,
            attack_feature=attribute_attack,
            attack_train_ratio=attack_train_ratio,
        )

        attacker.fit(x_train, y_train)

        y_attack, y_pred_attack = attacker.transform(x_test, y_test)

        return metric_fn(y_attack, y_pred_attack)


def attribute_attack_score(
    x_train: pd.DataFrame,
    x_test: pd.DataFrame,
    y_train: pd.Series,
    y_test: pd.Series,
    attribute_attack: str,
    attack_train_ratio: float = 0.5,
    **kargs,
) -> float:
    """
    Calculate the accuracy score for black box attribute attack. It is done as follows:
    - The attack attribute is removed from the training data.
    - The label is added as an input feature, and a machine learning model is trained.
    - The model is used to predict the removed attribute, and the prediction is compared with the actual value.

    Parameters
    ----------
    x_train: pd.DataFrame
        The training features.
    x_test: pd.DataFrame
        The testing features.
    y_train: pd.Series
        The training labels.
    y_test: pd.Series
        The testing labels.
    attribute_attack: str
        The attribute column in the x_train dataframe to attack.
    attack_train_ratio: float
        The ratio of the attack data to the training data.

    kargs: aditional attributes are passed to AttributeAttackScore class
    Returns
    -------
        float: The accuracy score for black box attribute attack.
    """

    bb = AttributeAttackScore()
    return bb(x_train, x_test, y_train, y_test, attribute_attack, attack_train_ratio, **kargs)



from fastapi import FastAPI, HTTPException
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel
from typing import List

import pandas as pd
import mlflow.pyfunc
import mlflow

# set the tracking server to be localhost
mlflow.set_tracking_uri(uri="http://127.0.0.1:8080")

# define the path to the MLflow model
logged_model = 'runs:/dede439c576a4d079f562e3d9ad7bb52/model_learning_fair_representation'

# load the MLflow model using mlflow.pyfunc.load_model()
loaded_model = mlflow.pyfunc.load_model(logged_model)

app = FastAPI()  # Initialize a FastAPI application

# define a Pydantic model for input data validation
class InputData(BaseModel):
    data: List[List[float]]
    columns: List[str]

# define a route and function for making predictions
@app.post('/predict')  # Define a route '/predict' that accepts HTTP POST requests
async def predict(input_data: InputData):
    try:
        input_df = pd.DataFrame(jsonable_encoder(input_data.data), columns=input_data.columns)

        # make predictions using the loaded model
        predictions = loaded_model.predict(input_df)  # Use the model to make predictions on input data

        # return predictions as a JSON response
        return {"predictions": predictions.tolist()}  # Convert predictions to a JSON response

    except Exception as e:
        # Handle exceptions (e.g., invalid input or model errors) and return an error message as a JSON response
        raise HTTPException(status_code=400, detail=str(e))


import jax
import jax.numpy as jnp
import optax
from flax import linen as nn
from flax.training import train_state


class ClassifierModel(nn.Module):
    features_dim: int
    hidden_size: int
    keep_prob: float

    def setup(self):
        self.encode = nn.Dense(self.hidden_size)
        self.decode = nn.Dense(1)

    @nn.compact
    def __call__(self, x, trainable=True):
        x = self.encode(x)
        x = nn.relu(x)
        x = nn.Dropout(rate=self.keep_prob, deterministic=not trainable)(x)
        x = self.decode(x)
        p = nn.sigmoid(x)
        return (p, x) if trainable else p


class AdversarialModel(nn.Module):
    def setup(self):
        self.c = self.param("c", jax.random.normal, (1,))
        self.hidden = nn.Dense(1)

    @nn.compact
    def __call__(self, x, y, trainable=True):
        s = nn.sigmoid((1 + jnp.abs(self.c)) * x)
        x = jnp.concatenate([s, s * y, s * (1.0 - y)], axis=1)
        x = self.hidden(x)
        p = nn.sigmoid(x)
        return (p, x) if trainable else p


class ADModel(nn.Module):
    classifier: nn.Module
    adversarial: nn.Module

    def __call__(self, x, y, trainable=True):
        y_prob, y_logits = self.classifier(x, trainable=trainable)
        z_prob, z_logits = self.adversarial(y_logits, y, trainable=trainable)
        return y_logits, z_logits


def train_step(cls_state, adv_state, batch, use_debias, adversary_loss_weight: float, rng):
    def loss_fn(classifier_params, adversarial_params, batch, rng):
        y = batch["y"].reshape([-1, 1])
        group = batch["group"].reshape([-1, 1])
        x = batch["X"]
        rngs = {"dropout": rng}
        _, y_logits = cls_state.apply_fn({"params": classifier_params}, x, trainable=True, rngs=rngs)
        _, z_logits = adv_state.apply_fn({"params": adversarial_params}, y_logits, y, trainable=True, rngs=rngs)

        loss_cls = optax.sigmoid_binary_cross_entropy(y_logits, y).mean()
        loss_adv = optax.sigmoid_binary_cross_entropy(z_logits, group).mean()
        if use_debias:
            return loss_cls - adversary_loss_weight * loss_adv, (loss_cls, loss_adv)
        return loss_cls, (loss_cls, None)

    def adv_loss_fn(adversarial_params, classifier_params, batch, rng):
        y = batch["y"].reshape([-1, 1])
        group = batch["group"].reshape([-1, 1])
        x = batch["X"]
        rngs = {"dropout": rng}
        _, y_logits = cls_state.apply_fn({"params": classifier_params}, x, trainable=True, rngs=rngs)
        _, z_logits = adv_state.apply_fn({"params": adversarial_params}, y_logits, y, trainable=True, rngs=rngs)

        loss_adv = optax.sigmoid_binary_cross_entropy(z_logits, group).mean()
        return loss_adv

    (loss, (loss_cls, loss_adv)), grads = jax.value_and_grad(loss_fn, argnums=(0), has_aux=True)(
        cls_state.params, adv_state.params, batch, rng
    )
    new_cls_state = cls_state.apply_gradients(grads=grads)

    loss_adv, grads = jax.value_and_grad(adv_loss_fn)(adv_state.params, new_cls_state.params, batch, rng)
    new_adv_state = adv_state.apply_gradients(grads=grads)

    return new_cls_state, new_adv_state, loss_cls, loss_adv


def create_train_state(rng, model, learning_rate, feature_dim):
    adversarial_params = model.adversarial.init(rng, jnp.ones([1, 1]), jnp.ones([1, 1]))["params"]
    classifier_params = model.classifier.init(rng, jnp.ones([1, feature_dim]))["params"]
    tx = optax.sgd(learning_rate, momentum=0.9)
    adversary_tx = optax.sgd(learning_rate, momentum=0.9)
    cls_state = train_state.TrainState.create(apply_fn=model.classifier.apply, params=classifier_params, tx=tx)
    adv_state = train_state.TrainState.create(
        apply_fn=model.adversarial.apply, params=adversarial_params, tx=adversary_tx
    )
    return cls_state, adv_state



from holisticai.explainability.metrics.local_importance._stability import (
    DataStability,
    FeatureStability,
    data_stability,
    feature_stability,
)

__all__ = ["DataStability", "FeatureStability", "data_stability", "feature_stability"]



import numpy as np
import pytest
from utils import (
    numeric_config,
    numeric_nan_config,
    str_config,
    unconventional_numeric_config,
)

from holisticai.bias.metrics import frequency_matrix, multiclass_statistical_parity

#######################
# 1 entrada + p_attr
#######################
# NaN Values
metrics_functions = [multiclass_statistical_parity, frequency_matrix]


@pytest.mark.parametrize("metric_function", metrics_functions)
def test_string_y_values(metric_function):
    """metric must accept string y values"""
    p_attr, y_pred, _ = str_config()
    metric_function(p_attr, y_pred)


@pytest.mark.parametrize("metric_function", metrics_functions)
def test_unvonventional_numeric_y(metric_function):
    """metric must accept unconventional numeric labels"""
    p_attr, y_pred, _ = unconventional_numeric_config()
    metric_function(p_attr, y_pred)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("metric_function", metrics_functions)
def test_y_pred_nan(metric_function):
    """metric must validate y_pred without nan values, the error must be ValueError and not KeyError"""
    p_attr, y_pred, _ = numeric_nan_config(y_pred_nan=True)
    metric_function(p_attr, y_pred)



from holisticai.robustness.attackers.classification.hop_skip_jump import HopSkipJump
from holisticai.robustness.attackers.classification.zeroth_order_optimization import ZooAttack

__all__ = ["HopSkipJump", "ZooAttack"]



from sklearn.naive_bayes import GaussianNB
from holisticai.datasets import load_dataset   
from holisticai.explainability.metrics import multiclass_explainability_metrics
from holisticai.explainability.metrics.global_importance import alpha_score, rank_alignment, position_parity, xai_ease_score
import numpy as np
import pytest

@pytest.fixture
def input_data():
    dataset = load_dataset('student_multiclass').select(list(range(1000)))
    dataset = dataset.train_test_split(test_size=0.2, random_state=42)
    train = dataset['test']
    test = dataset['test']

    model = GaussianNB()
    model.fit(train['X'], train['y'])
    return model, test


def get_multiclass_features(model, test):
    from holisticai.utils import MultiClassificationProxy
    from holisticai.utils.feature_importances import compute_permutation_feature_importance
    from holisticai.utils.inspection import compute_partial_dependence
    
    proxy = MultiClassificationProxy(predict=model.predict, predict_proba=model.predict_proba, classes=model.classes_)
    importances  = compute_permutation_feature_importance(proxy=proxy, X=test['X'], y=test['y'])
    ranked_importances = importances.top_alpha(0.8)
    partial_dependencies = compute_partial_dependence(test['X'], features=ranked_importances.feature_names, proxy=proxy)
    conditional_importances  = compute_permutation_feature_importance(proxy=proxy, X=test['X'], y=test['y'], conditional=True)
    return importances, ranked_importances, conditional_importances, partial_dependencies



@pytest.mark.skip(reason="Sometimes fails due to randomization of the dataset.")
@pytest.mark.parametrize("strategy, rank_alignment, position_parity, xai_ease_score, alpha_imp_score", [
    ("permutation", 0.4743650793650794, 0.129973544973545, 0.8833333333333333, 0.38461538461538464),
    ("surrogate",  0.2962962962962963, 0.2962962962962963, 0.7222222222222222, 0.11538461538461539)
])
def test_xai_multiclassification_metrics(strategy, rank_alignment, position_parity, xai_ease_score, alpha_imp_score, input_data):
    model, test = input_data
    metrics = multiclass_explainability_metrics(test['X'], test['y'], model.predict, model.predict_proba, classes=[0,1,2], strategy=strategy)
    assert np.isclose(metrics.loc['Rank Alignment'].value, rank_alignment)
    assert np.isclose(metrics.loc['Position Parity'].value, position_parity)
    assert np.isclose(metrics.loc['XAI Ease Score'].value, xai_ease_score)
    assert np.isclose(metrics.loc['Alpha Importance Score'].value, alpha_imp_score)

@pytest.mark.skip(reason="Sometimes fails due to randomization of the dataset.")
def test_xai_classification_metrics_separated(input_data):
    model, test = input_data
    
    importances, ranked_importances, conditional_importances, partial_dependencies = get_multiclass_features(model, test)
        
    value = rank_alignment(conditional_importances, ranked_importances)
    assert np.isclose(value, 0.4743650793650794)

    value = position_parity(conditional_importances, ranked_importances)
    assert np.isclose(value, 0.129973544973545)
    
    value = xai_ease_score(partial_dependencies, ranked_importances)
    assert np.isclose(value, 0.8833333333333333)

    value = alpha_score(importances)
    assert np.isclose(value, 0.38461538461538464)


import logging

import numpy as np
from sklearn.metrics.pairwise import pairwise_distances, pairwise_distances_argmin
from sklearn.utils import check_array, check_random_state
from sklearn.utils.extmath import stable_cumsum
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


class KMedoids:
    """k-medoids clustering.

    Read more in the :ref:`User Guide <k_medoids>`.

    Parameters
    ----------
    n_clusters : int, optional
        The number of clusters to form as well as the number of medoids to
        generate. Default is 8.

    metric : str, or callable, optional
        What distance metric to use. See :func:metrics.pairwise_distances
        metric can be 'precomputed', the user must then feed the fit method
        with a precomputed kernel matrix and not the design matrix X.
        Default is 'euclidean'. Possible values are 'euclidean', 'manhattan',

    method : str, optional
        Which algorithm to use. 'alternate' is faster while 'pam' is more accurate.
        Default is 'alternate'. Possible values are 'alternate' and 'pam'.

    init : str, optional
        Specify medoid initialization method. 'random' selects n_clusters
        elements from the dataset. 'heuristic' picks the n_clusters points
        with the smallest sum distance to every other point. 'k-medoids++'
        follows an approach based on k-means++_, and in general, gives initial
        medoids which are more separated than those generated by the other methods.
        'build' is a greedy initialization of the medoids used in the original PAM
        algorithm. Often 'build' is more efficient but slower than other
        initializations on big datasets and it is also very non-robust,
        if there are outliers in the dataset, use another initialization.
        Default is 'heuristic'. Possible values are 'random', 'heuristic', 'k-medoids++' and 'build'.

        .. _k-means++: https://theory.stanford.edu/~sergei/papers/kMeansPP-soda.pdf

    max_iter : int, optional
        Specify the maximum number of iterations when fitting. It can be zero in
        which case only the initialization is computed which may be suitable for
        large datasets when the initialization is sufficiently efficient
        (i.e. for 'build' init). Default is 300.

    random_state : int, RandomState instance or None, optional
        Specify random state for the random number generator. Used to
        initialise medoids when init='random'. Default is None.

    Attributes
    ----------
    cluster_centers_ : array, None if metric == 'precomputed'
        Cluster centers, i.e. medoids (elements from the original dataset), shape = (n_clusters, n_features)

    medoid_indices_ : array-like
        The indices of the medoid rows in X, shape = (n_clusters,)

    labels_ : array-like
        Labels of each point, shape = (n_samples,)

    inertia_ : float
        Sum of distances of samples to their closest cluster center.

    n_iter_ : int
        Number of iterations run.

    members : array-like
        The indices of the medoid rows in X, shape = (n_clusters,)

    centers : array-like
        The indices of the medoid rows in X, shape = (n_clusters,)

    Methods
    -------
    fit(X)
        Compute k-medoids clustering.

    predict(X)
        Predict the closest cluster each sample in X belongs to.

    transform(X)
        Transform X to a cluster-distance space.

    assign()
        Assigns every point in the dataset to the closest center.
    """

    def __init__(
        self,
        n_clusters=8,
        metric="euclidean",
        method="alternate",
        init="heuristic",
        max_iter=300,
        random_state=None,
    ):
        self.n_clusters = n_clusters
        self.metric = metric
        self.method = method
        self.init = init
        self.max_iter = max_iter
        self.random_state = random_state

    def _check_nonnegative_int(self, value, desc, strict=True):
        """
        Check if a value is a nonnegative integer.

        Parameters
        ----------
        value : int
            The value to check.
        desc : str
            The description of the value.
        strict : bool, optional
            If True, the value must be strictly positive. Default is True.

        Raises
        ------
        ValueError
            If the value is not a nonnegative integer.
        """
        negative = value is None or value <= 0 if strict else value is None or value < 0
        if negative or not isinstance(value, (int, np.integer)):
            raise ValueError(f"{desc} should be a nonnegative integer. " f"{value} was given")

    def _check_init_args(self):
        """
        Check the validity of the input arguments.

        Raises
        ------
        ValueError
            If the input arguments are invalid.
        """

        # Check n_clusters and max_iter
        self._check_nonnegative_int(self.n_clusters, "n_clusters")
        self._check_nonnegative_int(self.max_iter, "max_iter", False)

        # Check init
        init_methods = ["random", "heuristic", "k-medoids++", "build"]
        if self.init not in init_methods:
            raise ValueError("init needs to be one of " + "the following: " + f"{init_methods}")

    def fit(self, X, y=None, sample_weight=None):  # noqa: ARG002
        """
        Fit K-Medoids to the provided data.

        Parameters
        ----------
        X : array-like or sparse matrix
            The data to cluster. It must be an array of shape (n_samples, n_features) or (n_samples, n_samples)
            if metric == 'precomputed'

        y : Ignored
            Not used, present here for API consistency by convention.

        sample_weight : array-like, shape (n_samples,), optional
            The weights for each sample. If None, all samples are given equal weight.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        random_state_ = check_random_state(self.random_state)

        self._check_init_args()
        X = check_array(X, accept_sparse=["csr", "csc"])
        if self.n_clusters > X.shape[0]:
            raise ValueError(
                "The number of medoids (%d) must be less "
                "than the number of samples %d." % (self.n_clusters, X.shape[0])
            )

        D = pairwise_distances(X, metric=self.metric)
        medoid_idxs = self._initialize_medoids(D, self.n_clusters, random_state_)
        labels = None

        # Continue the algorithm as long as
        # the medoids keep changing and the maximum number
        # of iterations is not exceeded

        for n_iter_ in range(self.max_iter):
            old_medoid_idxs = np.copy(medoid_idxs)
            labels = np.argmin(D[medoid_idxs, :], axis=0)

            # Update medoids with the new cluster indices
            self._update_medoid_idxs_in_place(D, labels, medoid_idxs, sample_weight)

            if np.all(old_medoid_idxs == medoid_idxs):
                break
            if n_iter_ == self.max_iter - 1:
                logger.warning(
                    "Maximum number of iteration reached before "
                    "convergence. Consider increasing max_iter to "
                    "improve the fit."
                )

        # Set the resulting instance variables.
        if self.metric == "precomputed":
            self.cluster_centers_ = None
        else:
            self.cluster_centers_ = X[medoid_idxs]
        self.centers = medoid_idxs
        # Expose labels_ which are the assignments of
        # the training data to clusters
        self.labels_ = np.argmin(D[medoid_idxs, :], axis=0)
        self.members = self.labels_
        self.medoid_indices_ = medoid_idxs
        self.inertia_ = self._compute_inertia(self.transform(X))

        # Return self to enable method chaining
        return self

    def assign(self):
        """
        Assigns every point in the dataset to the closest center.

        Returns
        -------
        list
            The indices of the medoid rows in X
        """
        return list(enumerate(self.medoid_indices_))

    def _update_medoid_idxs_in_place(self, D, labels, medoid_idxs, sample_weight):
        """
        In-place update of the medoid indices

        Parameters
        ----------
        D : array-like
            The distance matrix between the data points, shape (n_samples, n_samples)
        labels : array-like
            The cluster labels for each data point, shape (n_samples,)
        medoid_idxs : array-like
            The indices of the medoid rows in X, shape = (n_clusters,)
        sample_weight : array-like, optional
            The weights for each sample. If None, all samples are given equal weight, shape (n_samples,)
        """

        # Update the medoids for each cluster
        for k in range(self.n_clusters):
            # Extract the distance matrix between the data points
            # inside the cluster k
            cluster_k_idxs = np.where(labels == k)[0]

            if len(cluster_k_idxs) == 0:
                logger.warning(
                    f"Cluster {k} is empty! "
                    f"self.labels_[self.medoid_indices_[{k}]] "
                    "may not be labeled with "
                    f"its corresponding cluster ({k})."
                )
                continue

            in_cluster_distances = D[cluster_k_idxs, cluster_k_idxs[:, np.newaxis]]

            # Calculate all costs from each point to all others in the cluster
            in_cluster_all_costs = np.sum(in_cluster_distances, axis=1)

            if sample_weight is not None:
                in_cluster_all_costs *= sample_weight[cluster_k_idxs]

            min_cost_idx = np.argmin(in_cluster_all_costs)
            min_cost = in_cluster_all_costs[min_cost_idx]
            curr_cost = in_cluster_all_costs[np.argmax(cluster_k_idxs == medoid_idxs[k])]

            # Adopt a new medoid if its distance is smaller then the current
            if min_cost < curr_cost:
                medoid_idxs[k] = cluster_k_idxs[min_cost_idx]

    def _compute_cost(self, D, medoid_idxs):
        """
        Compute the cose for a given configuration of the medoids.

        Parameters
        ----------
        D : array-like
            The distance matrix between the data points, shape (n_samples, n_samples)
        medoid_idxs : array-like
            The indices of the medoid rows in X, shape = (n_clusters,)

        Returns
        -------
        float
            The cost of the configuration.
        """
        return self._compute_inertia(D[:, medoid_idxs])

    def transform(self, X):
        """
        Transforms X to cluster-distance space.

        Parameters
        ----------
        X : array-like or sparse matrix
            Data to transform. It must be an array of shape (n_query, n_features) or (n_query, n_indexed)
            if metric == 'precomputed'

        Returns
        -------
        array-like or sparse matrix
            X transformed in the new space of distances to cluster centers, shape=(n_query, n_clusters)
        """
        X = check_array(X, accept_sparse=["csr", "csc"])

        if self.metric == "precomputed":
            check_is_fitted(self, "medoid_indices_")
            return X[:, self.medoid_indices_]

        check_is_fitted(self, "cluster_centers_")

        Y = self.cluster_centers_
        return pairwise_distances(X, Y=Y, metric=self.metric)

    def predict(self, X):
        """Predict the closest cluster for each sample in X.

        Parameters
        ----------
        X : array-like or sparse matrix
            New data to predict. It must be an array of shape (n_query, n_features) or (n_query, n_indexed)
            if metric == 'precomputed'

        Returns
        -------
        array-like
            Index of the cluster each sample belongs to, shape = (n_query,)
        """
        X = check_array(X, accept_sparse=["csr", "csc"])

        if self.metric == "precomputed":
            check_is_fitted(self, "medoid_indices_")
            return np.argmin(X[:, self.medoid_indices_], axis=1)

        check_is_fitted(self, "cluster_centers_")

        # Return data points to clusters based on which cluster assignment
        # yields the smallest distance
        return pairwise_distances_argmin(X, Y=self.cluster_centers_, metric=self.metric)

    def _compute_inertia(self, distances):
        """
        Compute inertia of new samples. Inertia is defined as the sum of the
        sample distances to closest cluster centers.

        Parameters
        ----------
        distances : array-like or sparse matrix
            Distances to cluster centers, shape=(n_samples, n_clusters)

        Returns
        -------
        float
            Sum of distances of samples to their closest cluster center.
        """

        # Define inertia as the sum of the sample-distances
        # to closest cluster centers
        inertia = np.sum(np.min(distances, axis=1))

        return inertia

    def _initialize_medoids(self, D, n_clusters, random_state_):
        """
        Select initial mediods when beginning clustering.

        Parameters
        ----------
        D : array-like
            The distance matrix between the data points, shape (n_samples, n_samples)
        n_clusters : int
            The number of clusters to form as well as the number of medoids to generate.
        random_state_ : RandomState instance
            The random number generator.

        Returns
        -------
        array-like
            The indices of the medoid rows in X, shape = (n_clusters,)
        """

        if self.init == "random":  # Random initialization
            # Pick random k medoids as the initial ones.
            medoids = random_state_.choice(len(D), n_clusters)
        elif self.init == "k-medoids++":
            medoids = self._kpp_init(D, n_clusters, random_state_)
        elif self.init == "heuristic":  # Initialization by heuristic
            # Pick K first data points that have the smallest sum distance
            # to every other point. These are the initial medoids.
            medoids = np.argpartition(np.sum(D, axis=1), n_clusters - 1)[:n_clusters]
        else:
            raise ValueError(f"init value '{self.init}' not recognized")

        return medoids

    def _kpp_init(self, D, n_clusters, random_state_, n_local_trials=None):
        """
        Init n_clusters seeds with a method similar to k-means++

        Parameters
        -----------
        D : array-like
            The distance matrix we will use to select medoid indices, shape (n_samples, n_samples)

        n_clusters : int
            The number of seeds to choose

        random_state : RandomState instance
            The generator used to initialize the centers.

        n_local_trials : integer, optional
            The number of seeding trials for each center (except the first),
            of which the one reducing inertia the most is greedily chosen.
            Set to None to make the number of trials depend logarithmically
            on the number of seeds (2+log(k)); this is the default.

        Returns
        -------
        array-like
            The indices of the medoid rows in X, shape = (n_clusters,)

        References
        ----------
        .. [1] Arthur, D. and Vassilvitskii, S.
        "k-means++: the advantages of careful seeding". ACM-SIAM symposium
        on Discrete algorithms. 2007

        Notes
        -----
        Version ported from http://www.stanford.edu/~darthur/kMeansppTest.zip,
        which is the implementation used in the aforementioned paper.
        """
        n_samples, _ = D.shape

        centers = np.empty(n_clusters, dtype=int)

        # Set the number of local seeding trials if none is given
        if n_local_trials is None:
            # This is what Arthur/Vassilvitskii tried, but did not report
            # specific results for other than mentioning in the conclusion
            # that it helped.
            n_local_trials = 2 + int(np.log(n_clusters))

        center_id = random_state_.randint(n_samples)
        centers[0] = center_id

        # Initialize list of closest distances and calculate current potential
        closest_dist_sq = D[centers[0], :] ** 2
        current_pot = closest_dist_sq.sum()

        # pick the remaining n_clusters-1 points
        for cluster_index in range(1, n_clusters):
            rand_vals = random_state_.random_sample(n_local_trials) * current_pot
            candidate_ids = np.searchsorted(stable_cumsum(closest_dist_sq), rand_vals)

            # Compute distances to center candidates
            distance_to_candidates = D[candidate_ids, :] ** 2

            # Decide which candidate is the best
            best_candidate = None
            best_pot = None
            best_dist_sq = None
            for trial in range(n_local_trials):
                # Compute potential when including center candidate
                new_dist_sq = np.minimum(closest_dist_sq, distance_to_candidates[trial])
                new_pot = new_dist_sq.sum()

                # Store result if it is the best local trial so far
                if (best_candidate is None) or (new_pot < best_pot):
                    best_candidate = candidate_ids[trial]
                    best_pot = new_pot
                    best_dist_sq = new_dist_sq

            centers[cluster_index] = best_candidate
            current_pot = best_pot
            closest_dist_sq = best_dist_sq

        return centers



from __future__ import annotations

import itertools
import sys
from typing import Literal

import holisticai.bias.metrics as bias_metrics
import numpy as np
from holisticai.utils.transformers.bias import BMPostprocessing as BMPost
from sklearn.metrics import balanced_accuracy_score


def statistical_parity(group_a, group_b, y_pred, _):
    return bias_metrics.statistical_parity(group_a, group_b, y_pred)


class RejectOptionClassification(BMPost):
    """
    Reject option classification gives favorable outcomes (y=1) to unpriviliged groups and unfavorable outcomes (y=0) to\
    priviliged groups in a confidence band around the decision boundary with the highest uncertainty.

    Parameters
    ----------
    low_class_thresh : float
        Smallest classification threshold to use in the optimization.
        Should be between 0. and 1.
    low_class_thresh : float
        Smallest classification threshold to use in the optimization.
        Should be between 0. and 1.
    high_class_thresh : float
        Highest classification threshold to use in the optimization.
        Should be between 0. and 1.
    num_class_thresh : int
        Number of classification thresholds between low_class_thresh and high_class_thresh for the optimization
        search. Should be > 0.
    num_ROC_margin : int
        Number of relevant ROC margins to be used in the optimization search. Should be > 0.
    metric_name : str
        Name of the metric to use for the optimization. Allowed options are:
        "Statistical parity difference",
        "Average odds difference",
        "Equal opportunity difference".
    metric_ub : float
        Upper bound of constraint on the metric value
    metric_lb : float
        Lower bound of constraint on the metric value
    verbose : int
        If >0, will show progress percentage.

    References
    ----------
        .. [1] Kamiran, Faisal, Asim Karim, and Xiangliang Zhang. "Decision theory for discrimination-aware classification."\
        2012 IEEE 12th International Conference on Data Mining. IEEE, 2012.
    """

    ALLOWED_METRICS = Literal[
        "Statistical parity difference",
        "Average odds difference",
        "Equal opportunity difference",
    ]

    def __init__(
        self,
        low_class_thresh: float | None = 0.01,
        high_class_thresh: float | None = 0.99,
        num_class_thresh: int | None = 100,
        num_ROC_margin: int | None = 50,
        metric_name: str | None = "Statistical parity difference",
        metric_ub: float | None = 0.05,
        metric_lb: float | None = -0.05,
        num_workers: int | None = 4,
        verbose: int | None = 0,
    ):
        super().__init__()

        allowed_metrics = [
            "Statistical parity difference",
            "Average odds difference",
            "Equal opportunity difference",
        ]

        self.low_class_thresh = low_class_thresh
        self.high_class_thresh = high_class_thresh
        self.num_class_thresh = num_class_thresh
        self.num_ROC_margin = num_ROC_margin
        self.metric_name = metric_name
        self.metric_ub = metric_ub
        self.metric_lb = metric_lb
        self.num_workers = num_workers
        self.verbose = verbose

        self.classification_threshold = None
        self.ROC_margin = None

        if (
            (self.low_class_thresh < 0.0)
            or (self.low_class_thresh > 1.0)
            or (self.high_class_thresh < 0.0)
            or (self.high_class_thresh > 1.0)
            or (self.low_class_thresh >= self.high_class_thresh)
            or (self.num_class_thresh < 1)
            or (self.num_ROC_margin < 1)
        ):
            msg = "Input parameter values out of bounds"
            raise ValueError(msg)

        if metric_name not in allowed_metrics:
            msg = "metric name not in the list of allowed metrics"
            raise ValueError(msg)

    def fit(
        self,
        y: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Compute parameters for reject option classification strategy.

        Description
        ----------
        Evaluate the likelihood with different thresholds. Select the threshold and ROC margin \
        with the best fair metric value and highest balanced accuracy.

        Parameters
        ----------
        y : array-like
            Target vector (nb_examples,)
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
            Self
        """

        params = self._load_data(y=y, y_proba=y_proba, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y = params["y"]
        likelihoods = params["y_score"]
        y_proba = params["y_proba"]

        class_thresholds = np.linspace(self.low_class_thresh, self.high_class_thresh, self.num_class_thresh)

        if self.metric_name == "Statistical parity difference":
            fair_metric = statistical_parity

        elif self.metric_name == "Average odds difference":
            fair_metric = bias_metrics.average_odds_diff

        elif self.metric_name == "Equal opportunity difference":
            fair_metric = bias_metrics.equal_opportunity_diff

        else:
            msg = "metric name not in the list of allowed metrics"
            raise ValueError(msg)

        args_iterator = [
            (
                fair_metric,
                y,
                likelihoods,
                group_a,
                group_b,
                class_threshold,
                self.num_ROC_margin,
            )
            for class_threshold in class_thresholds
        ]

        # Serial
        # configurations = []
        # self.num_cases = len(args_iterator)
        # for i, argv in enumerate(args_iterator):
        #    configurations.append(_evaluate_threshold(*argv))
        #    self._log_progres(i)

        # Pool
        from multiprocessing import Pool

        with Pool(self.num_workers) as p:
            configurations = list(p.starmap(_evaluate_threshold, args_iterator))

        # Joblib
        # from joblib import Parallel, delayed
        # configurations = Parallel(n_jobs=-1)(
        #    delayed(_evaluate_threshold)(*args) for args in args_iterator
        # )

        configurations = list(itertools.chain.from_iterable(configurations))

        selected_configurations = list(
            filter(
                lambda c: (c["fair_score"] >= self.metric_lb) and (c["fair_score"] <= self.metric_ub),
                configurations,
            )
        )

        if any(selected_configurations):
            best_config = max(selected_configurations, key=lambda c: c["balanced_accurracy"])
        else:
            best_config = min(configurations, key=lambda c: abs(c["fair_score"]))

        self.ROC_margin = best_config["roc_margin"]
        self.classification_threshold = best_config["class_thresh"]

        return self

    def transform(
        self,
        y_pred: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Obtain fair predictions using the ROC method.

        Description
        ----------
        Predict the output using the fitted threshold and ROC margin.

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examples,)
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability\
            estimates must sum to 1 across the possible classes and each matrix value\
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dict
            A dictionary of new predictions
        """
        params = self._load_data(y_pred=y_pred, y_proba=y_proba, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        y_score = params["y_score"]

        new_y_pred = predict(
            y_score > self.classification_threshold,
            y_score,
            group_a,
            group_b,
            self.classification_threshold,
            self.ROC_margin,
        )
        new_y_score = y_score.copy()

        return {
            "y_pred": new_y_pred,
            "y_score": new_y_score,
        }

    def _log_progres(self, i):
        if self.verbose > 0:
            sys.stdout.write(f"\rProgress: {i+1}/{self.num_cases}")
            sys.stdout.flush()


def _evaluate_threshold(fair_metric, labels, likelihoods, group_a, group_b, class_thresh, num_roc_margin):
    base_predictions = np.where(likelihoods > class_thresh, 1, 0)
    high_roc_margin = class_thresh if class_thresh <= 0.5 else 1.0 - class_thresh
    roc_margins = np.linspace(0.0, high_roc_margin, num_roc_margin)
    configurations = []
    for roc_margin in roc_margins:
        prediction = predict(base_predictions, likelihoods, group_a, group_b, class_thresh, roc_margin)
        configuration = {
            "class_thresh": class_thresh,
            "roc_margin": roc_margin,
            "balanced_accurracy": balanced_accuracy_score(labels, prediction),
            "fair_score": fair_metric(group_b, group_a, prediction, labels),
        }
        configurations.append(configuration)
    return configurations


def predict(predictions, likelihoods, group_a, group_b, threshold, roc_margin):
    """
    Predict the output using the ROC method.

    Description
    ----------
    Predict the output using the fitted threshold and ROC margin.

    Parameters
    ----------
    predictions : array-like
        Predicted vector (nb_examples,)
    likelihoods : matrix-like
        Predicted probability matrix (num_examples, num_classes). The probability\
        estimates must sum to 1 across the possible classes and each matrix value\
        must be in the interval [0,1].
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    threshold : float
        float value to discriminate between 0 and 1
    roc_margin : float
        float value to determine the margin around the decision boundary

    Returns
    -------
    array-like
        A new array of predictions
    """
    # Indices of critical region around the classification boundary
    upper_threshold = threshold + roc_margin
    lower_threshold = threshold - roc_margin
    crit_region_inds = np.logical_and(likelihoods <= upper_threshold, likelihoods > lower_threshold)

    # New, fairer labels
    new_predictions = predictions.copy()
    new_predictions[np.logical_and(crit_region_inds, group_a)] = 1
    new_predictions[np.logical_and(crit_region_inds, group_b)] = 0
    return new_predictions



import numpy as np
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler


def append_if_not_empty(original_array, array_to_append):
    if len(array_to_append) > 0:
        original_array = np.append(original_array, array_to_append)
    return original_array


class BlackBoxAttack:
    def __init__(self, attacker_estimator, attack_feature, attack_train_ratio=0.5):
        self.attack_feature = attack_feature
        self.attack_train_ratio = attack_train_ratio
        self.attacker_estimator = attacker_estimator

    def create_preprocessor(self, X):
        categorical_features = X.select_dtypes(include=["category"]).columns
        numerical_fatures = X.select_dtypes(exclude=["category"]).columns

        # Create transformers for numerical and categorical features
        numeric_transformer = Pipeline(steps=[("scaler", StandardScaler())])
        categorical_transformer = Pipeline(steps=[("onehot", OneHotEncoder(handle_unknown="ignore"))])

        # Combine transformers into a preprocessor using ColumnTransformer
        return ColumnTransformer(
            transformers=[
                ("num", numeric_transformer, numerical_fatures),
                ("cat", categorical_transformer, categorical_features),
            ]
        )

    def fit(self, X, y):
        """

        Description
        -----------
        The black-box attack basically trains an additional classifier (called the attack model) to predict the attacked feature's value from the remaining n-1
        features as well as the original (attacked) model's predictions.

        Parameters
        ----------

        X : pandas Dataframe
            input matrix

        y : numpy array
            Target vector of original model

        Returns
        -------

        np.ndarray: Predicted output per sample.
        """

        categorical_features = []
        y_train_attack = X[self.attack_feature]
        X_train_attack = X.drop(columns=[self.attack_feature])
        X_train_attack["label"] = y
        if y_train_attack.dtype == "category":
            categorical_features.append("label")
        # Create transformers for numerical and categorical features

        preprocessor = self.create_preprocessor(X_train_attack)

        attacker = Pipeline(steps=[("preprocessor", preprocessor), ("estimator", self.attacker_estimator)])

        attack_train_size = int(X_train_attack.shape[0] * self.attack_train_ratio)
        attacker.fit(X_train_attack[:attack_train_size], y_train_attack[:attack_train_size])
        self.attacker = attacker
        return self

    def transform(self, X, y):
        y_attack = X[self.attack_feature]
        X_test_attack = X.drop(columns=[self.attack_feature])
        X_test_attack["label"] = y
        y_pred_attack = self._predict(X_test_attack)
        return y_attack, y_pred_attack

    def _predict(self, x_attack):
        return self.attacker.predict(x_attack)


def classification_security_features(X, y, attacker, attack_feature):
    if attacker == "black_box":
        attacker = BlackBoxAttack(learning_task="classification", attack_feature=attack_feature, attack_train_ratio=0.5)
        attacker.fit(X, y)
        return attacker
    raise ValueError("Invalid attacker type. Please choose from 'black_box'")




import warnings
warnings.filterwarnings("ignore")

from holisticai.datasets import load_dataset
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LogisticRegression
from holisticai.robustness.attackers import HopSkipJump, ZooAttack
from holisticai.utils import BinaryClassificationProxy
from holisticai.robustness.attackers.classification.commons import format_function_predict_proba
from holisticai.robustness.metrics import adversarial_accuracy, empirical_robustness
from sklearn.metrics import accuracy_score
import pytest

SHARD_SIZE=60

@pytest.fixture
def categorical_dataset():
    dataset = load_dataset("adult", protected_attribute='race') # x,y,group_a,group_b
    dataset = dataset.groupby(['y','group_a']).sample(SHARD_SIZE, random_state=42) # 0-ga | 0-gb  | 1-ga | 1-gb
    train_test = dataset.train_test_split(test_size=0.02, stratify=dataset['y'], random_state=0)
    train = train_test['train']
    test = train_test['test']
    correlations = train['X'].corrwith(train['y']).sort_values(ascending=False)
    top_10_features = correlations.head(10).index.tolist()
    train['X'] = train['X'][top_10_features]
    test['X'] = test['X'][top_10_features]
    return train, test


def test_hsj(categorical_dataset):
    train, test = categorical_dataset

    # Standardize data and fit model
    pipe = Pipeline([('scaler', StandardScaler()), ('lr', LogisticRegression())])
    pipe.fit(train['X'], train['y'])

    y_pred = pipe.predict(test['X'])
    baseline_accuracy = accuracy_score(test['y'], y_pred)

    proxy = BinaryClassificationProxy(predict=pipe.predict, predict_proba=pipe.predict_proba, classes=[0, 1])

    hsj_attacker = HopSkipJump(name="HSJ", predictor=proxy.predict)

    hsj_adv_x = hsj_attacker.generate(test['X'])
    y_adv_pred = proxy.predict(hsj_adv_x)

    hsj_accuracy = adversarial_accuracy(test['y'], y_pred, y_adv_pred)
    hsj_robustness = empirical_robustness(test['X'], hsj_adv_x, y_pred, y_adv_pred, norm=2)


    # This test should pass when both dataframes are different
    assert not hsj_adv_x.equals(test['X'])
    # This test should pass when both accuracies are different
    assert hsj_accuracy != baseline_accuracy
    # This test should pass when the robustness is greater than 0
    assert hsj_robustness > 0

def test_zoo(categorical_dataset):
    train, test = categorical_dataset

    # Standardize data and fit model
    pipe = Pipeline([('scaler', StandardScaler()), ('lr', LogisticRegression())])
    pipe.fit(train['X'], train['y'])

    y_pred = pipe.predict(test['X'])
    baseline_accuracy = accuracy_score(test['y'], y_pred)

    proxy = BinaryClassificationProxy(predict=pipe.predict, predict_proba=pipe.predict_proba, classes=[0, 1])

    zoo_attacker = ZooAttack(name="Zoo", proxy=proxy)

    zoo_adv_x = zoo_attacker.generate(test['X'])

    y_adv_pred = proxy.predict(zoo_adv_x)
    zoo_accuracy = adversarial_accuracy(test['y'], y_pred, y_adv_pred)
    zoo_robustness = empirical_robustness(test['X'], zoo_adv_x, y_pred, y_adv_pred, norm=2)

    # This test should pass when both dataframes are different
    assert not zoo_adv_x.equals(test['X'])
    # This test should pass when both accuracies are different
    assert zoo_accuracy != baseline_accuracy
    # This test should pass when the robustness is greater than 0
    assert zoo_robustness > 0



import warnings

from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import KMeans

from holisticai.pipeline import Pipeline
from holisticai.bias.metrics import classification_bias_metrics, regression_bias_metrics, multiclass_bias_metrics, clustering_bias_metrics
from holisticai.datasets.synthetic.recruitment import generate_rankings
from holisticai.bias.mitigation.postprocessing.debiasing_exposure.algorithm_utils import exposure_metric

from tests.bias.mitigation.utils import (
    check_results,
    categorical_dataset,
    regression_dataset,
    multiclass_dataset,
    clustering_dataset
)
import pytest

warnings.filterwarnings("ignore")

seed = 42

from holisticai.bias.mitigation import MITIGATOR_NAME
from holisticai.bias.mitigation import CorrelationRemover
from holisticai.bias.mitigation import Reweighing
from holisticai.bias.mitigation import LearningFairRepresentation
from holisticai.bias.mitigation import FairletClusteringPreprocessing
from holisticai.bias.mitigation import DisparateImpactRemoverRS

def get_preprocessor(mitigator_name : MITIGATOR_NAME = "CorrelationRemover", parameters: dict = {}):
    if mitigator_name == "CorrelationRemover":
        return CorrelationRemover(**parameters)
    elif mitigator_name == "Reweighing":
        return Reweighing(**parameters)
    elif mitigator_name == "LearningFairRepresentation":
        return LearningFairRepresentation(**parameters)
    elif mitigator_name == "FairletClusteringPreprocessing":
        return FairletClusteringPreprocessing(**parameters)
    elif mitigator_name == "DisparateImpactRemoverRS":
        return DisparateImpactRemoverRS(**parameters)
    else:
        raise NotImplementedError


@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params, extra_model_fit_params", [
    ("CorrelationRemover", {"alpha":1.0}, ['X','group_a','group_b'], []),
    ("Reweighing", {}, ['X','y','group_a','group_b'], ['sample_weight']),
])
def test_multiclass_preprocessor(mitigator_name, mitigator_params, fit_params, extra_model_fit_params, multiclass_dataset):
    metrics1 = run_preprocessing_categorical(multiclass_dataset, multiclass_bias_metrics, LogisticRegression, mitigator_name, fit_params, extra_model_fit_params, mitigator_params, is_multiclass=True)
    metrics2 = run_preprocessing_catergorical_pipeline(multiclass_dataset, multiclass_bias_metrics, LogisticRegression, mitigator_name, mitigator_params, is_multiclass=True)
    check_results(metrics1, metrics2)


@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params, extra_model_fit_params", [
    ("CorrelationRemover", {"alpha":1.0}, ['X','group_a','group_b'], []),
    ("Reweighing", {}, ['X','y','group_a','group_b'], ['sample_weight']),
    ("LearningFairRepresentation", {}, ['X','y','group_a','group_b'], [])
])
def test_categorical_preprocessor(mitigator_name, mitigator_params, fit_params, extra_model_fit_params, categorical_dataset):
    metrics1 = run_preprocessing_categorical(categorical_dataset, classification_bias_metrics, LogisticRegression, mitigator_name, fit_params, extra_model_fit_params, mitigator_params)
    metrics2 = run_preprocessing_catergorical_pipeline(categorical_dataset, classification_bias_metrics, LogisticRegression, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)


@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params", [
    ("CorrelationRemover", {"alpha":1.0}, ['X','group_a','group_b']),
])
def test_regression_preprocessor(mitigator_name, mitigator_params, fit_params, regression_dataset):
    metrics1 = run_preprocessing_regression(regression_dataset, LinearRegression, mitigator_name, fit_params, mitigator_params)
    metrics2 = run_preprocessing_regression_pipeline(regression_dataset, LinearRegression, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)

@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params", [
    ("FairletClusteringPreprocessing", {"seed":42, "p":1, "q":3}, ['X','group_a','group_b']),
])
def test_clustering_preprocessor(mitigator_name, mitigator_params, fit_params, clustering_dataset):
    metrics1 = run_preprocessing_clustering(clustering_dataset, KMeans, mitigator_name, fit_params, mitigator_params)
    metrics2 = run_preprocessing_clustering_pipeline(clustering_dataset, KMeans, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)

@pytest.mark.parametrize("mitigator_name, mitigator_params", [
    ("DisparateImpactRemoverRS", {'query_col': 'X', 'group_col': 'protected', 'score_col': 'score', 'repair_level':1}),
])
def test_recsys_preprocessor(mitigator_name, mitigator_params):
    import pandas as pd
    metrics1 = run_preprocessing_recsys(exposure_metric, mitigator_name, mitigator_params)
    metrics2 = pd.DataFrame(columns=['Value'], data=[1.0037613239703396, 0.001944], index=['exposure_ratio', 'exposure difference'])
    check_results(metrics1, metrics2)


def run_preprocessing_categorical(dataset, bias_metrics, estimator_class, mitigator_name, preprocessor_fit_param_names, extra_model_fit_params,  mitigator_params, is_multiclass=False):
    train = dataset['train']
    test = dataset['test']

    pre = get_preprocessor(mitigator_name, mitigator_params)
    xt = pre.fit_transform(**{p:train[p] for p in preprocessor_fit_param_names})

    model = estimator_class(random_state=42)
    model_fit_parmams = {'X':xt, 'y':train['y']}
    
    if 'sample_weight' in extra_model_fit_params:
        model_fit_parmams.update({'sample_weight': pre.sample_weight})
    model.fit(**model_fit_parmams)

    xt = pre.transform(test['X'], group_a=test['group_a'], group_b=test['group_b'])
    y_pred = model.predict(xt)   
    
    if is_multiclass:
        return bias_metrics(test['group_a'],  y_pred, test['y'])
    else:
        return bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_preprocessing_catergorical_pipeline(dataset, bias_metrics, estimator_class, mitigator_name, mitigator_params, is_multiclass=False):
    train = dataset['train']
    test = dataset['test']

    pipeline = Pipeline(
        steps=[
            ("bm_preprocessing", get_preprocessor(mitigator_name, mitigator_params)),
            ("estimator", estimator_class(random_state=42)),
        ]
    )

    pipeline.fit(train['X'], train['y'], bm__group_a=train['group_a'], bm__group_b=train['group_b'])
    y_pred = pipeline.predict(test['X'], bm__group_a=test['group_a'], bm__group_b=test['group_b'])
    
    if is_multiclass:
        return bias_metrics(test['group_a'],  y_pred, test['y'])
    else:
        return bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_preprocessing_regression(dataset, estimator_class, mitigator_name, preprocessor_fit_param_names,  mitigator_params):
    train = dataset['train']
    test = dataset['test']

    pre = get_preprocessor(mitigator_name, mitigator_params)
    xt = pre.fit_transform(**{p:train[p] for p in preprocessor_fit_param_names})

    model = estimator_class()
    model_fit_parmams = {'X':xt, 'y':train['y']}
    model.fit(**model_fit_parmams)

    xt = pre.transform(test['X'], group_a=test['group_a'], group_b=test['group_b'])
    y_pred = model.predict(xt)   
    
    return regression_bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_preprocessing_regression_pipeline(dataset, estimator_class, mitigator_name, mitigator_params):
    train = dataset['train']
    test = dataset['test']

    pipeline = Pipeline(
        steps=[
            ("bm_preprocessing", get_preprocessor(mitigator_name, mitigator_params)),
            ("estimator", estimator_class()),
        ]
    )

    pipeline.fit(train['X'], train['y'], bm__group_a=train['group_a'], bm__group_b=train['group_b'])
    y_pred = pipeline.predict(test['X'], bm__group_a=test['group_a'], bm__group_b=test['group_b'])
    
    return regression_bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])

def run_preprocessing_clustering(dataset, estimator_class, mitigator_name, fit_params_name, mitigator_params):
    train = dataset['train']

    pre = get_preprocessor(mitigator_name, mitigator_params)
    fit_params = {p:train[p] for p in fit_params_name}
    xt = pre.fit_transform(**fit_params)

    model = estimator_class(n_clusters=4, random_state=42)
    model.fit(xt)

    y_pred = model.predict(xt)
    centroids = model.cluster_centers_
    return clustering_bias_metrics(train['group_a'], train['group_b'], y_pred, train['X'], centroids)

def run_preprocessing_clustering_pipeline(dataset, estimator_class, mitigator_name, mitigator_params):
    train = dataset['train']

    pipeline = Pipeline(
        steps=[
            ("bm_preprocessing", get_preprocessor(mitigator_name, mitigator_params)),
            ("estimator", estimator_class(n_clusters=4, random_state=42)),
        ]
    )

    fit_params_dict = {
        "bm__group_a": train['group_a'], 
        "bm__group_b": train['group_b']
        }
    
    pipeline.fit(train['X'], **fit_params_dict)

    y_pred = pipeline.predict(train['X'])
    centroids = pipeline.named_steps['estimator'].cluster_centers_

    return clustering_bias_metrics(train['group_a'], train['group_b'], y_pred, train['X'], centroids)

def run_preprocessing_recsys(bias_metrics, mitigator_name, mitigator_params):
    rankings = generate_rankings(M=1000, k=20, p=0.25, return_p_attr=False)
    pre = get_preprocessor(mitigator_name, mitigator_params)
    rankings = pre.transform(rankings)

    return bias_metrics(rankings, group_col='protected', query_col='X', score_col='score')



import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
import os
import sys
import shutil

def run_notebook(notebook_path):
    with open(notebook_path) as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600, kernel_name='python3')
        try:
            ep.preprocess(nb, {'metadata': {'path': os.path.dirname(notebook_path)}})
            with open(notebook_path, 'wt') as f:
                nbformat.write(nb, f)
            print(f"Executed: {notebook_path}")
            print(f"Output: {notebook_path}")
        except Exception as e:
            print(f"Error executing the notebook {notebook_path}: {e}")

def run_all_notebooks(folder_path):
    for root, _, files in os.walk(folder_path):
        for file in files:
            if file.endswith('.ipynb'):
                notebook_path = os.path.join(root, file)
                print(f"Running notebook: {notebook_path}")
                run_notebook(notebook_path)

def copy_folder(origen, destino):
    try:
        if not os.path.exists(destino):
            os.makedirs(destino)
        shutil.copytree(origen, destino, dirs_exist_ok=True)
        #run_all_notebooks(destino)
        print(f"Folder copied from {origen} to {destino} sucessfully.")
    except Exception as e:
        print(f"Error when trying to copy folder: {e}")


if __name__ == '__main__':
    # Run all the tutorials
    work_dir = os.getcwd()
    src_path = os.path.join(work_dir,'src')
    tutorials_path = os.path.join(work_dir,'tutorials')
    docs_path = os.path.join(work_dir,'docs','source')
    sys.path.insert(0, src_path)

    verticals = ['bias', 'datasets', 'explainability', 'robustness', 'security']  

    for vertical in verticals:
        path = os.path.join(tutorials_path, vertical)
        copy_folder(path, os.path.join(docs_path, 'gallery', 'tutorials', vertical))


from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
from numpy.random import RandomState

from holisticai.datasets import Dataset
from holisticai.utils import LocalImportances, ModelProxy
from holisticai.utils.feature_importances import group_samples_by_learning_task


def compute_shap_feature_importance(
    X: pd.DataFrame,
    y: pd.Series,
    proxy: ModelProxy,
    max_samples: Union[int, None] = None,
    random_state: Union[RandomState, None] = None,
) -> LocalImportances:
    if random_state is None:
        random_state = RandomState(42)

    if max_samples is None:
        max_samples = -1

    ds = Dataset(X=X, y=y)
    if max_samples > 0:
        ds = ds.sample(n=max_samples, random_state=random_state)

    pfi = SHAPImportanceCalculator()
    data = pfi.compute_importances(ds, proxy)
    condition = group_samples_by_learning_task(ds["y"], proxy.learning_task, return_group_mask=True)
    local_importances = LocalImportances(data=data, cond=condition)
    return local_importances


class SHAPImportanceCalculator:
    importance_type: str = "local"

    def initialize_explainer(self, X: pd.DataFrame, proxy: ModelProxy):
        try:
            import shap
        except ImportError:
            raise ImportError("SHAP is not installed. Please install it using 'pip install shap'") from None

        masker = shap.maskers.Independent(X)
        self.explainer = shap.Explainer(proxy.predict, masker=masker)

    def compute_importances(self, ds: Dataset, proxy: ModelProxy):
        X = ds["X"].astype(np.float64)
        self.initialize_explainer(X, proxy)
        shap_values = self.explainer(X)

        data = np.abs(shap_values.values)
        data = data / data.sum(axis=1)[:, None]
        return pd.DataFrame(data=data, columns=X.columns)



from holisticai.utils import Importances
from matplotlib import pyplot as plt


def plot_feature_importance(feature_importance: Importances, alpha=0.8, top_n=20, ax=None):
    """
    Bar plot of ranked feature importance.

    Parameters
    ----------
    feature_importance: Importances
        The feature importance data.
    top_n: (int, optional)
        The number of top features to display. Defaults to 20.
    alpha: (float, optional)
        Percentage of importance to consider as top features. Defaults to 0.8.
    ax: (matplotlib.axes.Axes, optional)
        The matplotlib axes to plot on. If not provided, a new figure and axes will be created.

    Returns
    -------
        matplotlib.axes.Axes: The matplotlib axes object containing the plot.

    Example
    -------
    >>> feature_importance = Importances(
    ...     values=np.array([0.1, 0.2, 0.3, 0.4]), feature_names=["A", "B", "C", "D"]
    ... )
    >>> plot_feature_importance(feature_importance)

    The plot should look like this:

    .. image:: /_static/images/xai_plot_feature_importance.png
        :alt: Plot Feature Importance

    """

    ranked_feature_importance = feature_importance.top_alpha(alpha=alpha)
    ranked_feature_importance = ranked_feature_importance.as_dataframe().set_index("Variable")
    feature_importances = feature_importance.as_dataframe().set_index("Variable")

    feature_importances.loc[:, "color"] = "#21918C"
    feature_importances.loc[ranked_feature_importance.index, "color"] = "#440154"
    feature_importances.reset_index(inplace=True, drop=False)

    top_n = min(top_n, len(feature_importances))
    top_features = feature_importances.sort_values(by="Importance", ascending=True).tail(top_n)
    if ax is None:
        _, ax = plt.subplots(1, 1)
    ax = top_features.plot(kind="barh", x="Variable", y="Importance", color=top_features["color"], legend=False, ax=ax)
    ax.axhline(y=len(top_features) - len(ranked_feature_importance) - 0.5, color="red", linestyle="--", linewidth=2)
    ax.grid()
    ax.set_xlabel("Importance")
    ax.set_ylabel("Features")

    if hasattr(feature_importance, "strategy"):
        ax.set_title(f"{feature_importance.strategy.title()} Feature Importance")
    else:
        ax.set_title("Feature Importance")

    return ax



import logging
import sys
import time

import numpy as np

logger = logging.getLogger(__name__)


class GAHiperparameters:
    def __init__(self, **kargs):
        self.max_num_iteration = kargs.get("max_num_iteration", None)
        self.population_size = kargs.get("population_size", 100)
        self.mutation_probability = kargs.get("mutation_probability", 0.1)
        self.elit_ratio = kargs.get("elit_ratio", 0.01)
        self.crossover_probability = kargs.get("crossover_probability", 0.5)
        self.parents_portion = kargs.get("parents_portion", 0.3)
        self.crossover_type = kargs.get("crossover_type", "uniform")
        self.max_iteration_without_improv = kargs.get("max_iteration_without_improv", None)


class GeneticAlgorithm:
    """
    Genetic Algorithm (Elitist version)
    Implementation of elitist genetic algorithm for solving problems with
    continuous, integers, or mixed variables.
    """

    def __init__(
        self,
        function,
        dimension: int,
        variable_type: str = "bool",
        variable_boundaries: np.ndarray = None,
        variable_type_mixed: np.ndarray = None,
        algorithm_parameters: GAHiperparameters = None,
        verbose: int = 0,
    ):
        """
        Parameters
        ----------

        function : <Callable>
        The given objective function to be minimized.

        dimension : int
        The number of decision variables.

        variable_type: str
            - 'bool' if all variables are Boolean;
            - 'int' if all variables are integer; and
            - 'real' if all variables are real value or continuous

        for mixed type see @param variable_type_mixed.

        variable_boundaries: array-like
        None if variable_type is 'bool'; otherwise provide an array of tuples
        of length two as boundaries for each variable;
        the length of the array must be equal dimension. For example,
        np.array([0,100],[0,200]) determines lower boundary 0 and upper boundary 100 for first
        and upper boundary 200 for second variable where dimension is 2.

        variable_type_mixed: array-like
        None if all variables have the same type; otherwise this can be used to
        specify the type of each variable separately. For example if the first
        variable is integer but the second one is real the input is:
        np.array(['int'],['real']). NOTE: it does not accept 'bool'. If variable
        type is Boolean use 'int' and provide a boundary as [0,1]
        in variable_boundaries. Also if variable_type_mixed is applied,
        variable_boundaries has to be defined.

        algorithm_parameters: GAArguments
        class containing GA algoirhtm hiperparameters

        verbose: int
        If >0, log information

        """
        self.verbose = verbose

        # input function
        assert callable(function), "function must be callable"
        self.f = function
        self.dim = int(dimension)

        # input variable type
        if variable_type not in ("bool", "int", "real"):
            msg = "variable_type must be 'bool', 'int', or 'real'"
            raise ValueError(msg)

        # input variables' type (MIXED)
        if variable_type_mixed is None:
            if variable_type == "real":
                self.var_type = np.array([["real"]] * self.dim)
            else:
                self.var_type = np.array([["int"]] * self.dim)
        else:
            if type(variable_type_mixed).__module__ != "numpy":
                raise ValueError("variable_type must be numpy array")
            if len(variable_type_mixed) != self.dim:
                raise ValueError("variable_type must have a length equal to dimension.")

            for i in variable_type_mixed:
                assert i in ("real", "int"), (
                    "\n variable_type_mixed is either 'int' or 'real' "
                    "ex:['int','real','real']"
                    "\n for 'boolean' use 'int' and specify boundary as [0,1]"
                )

            self.var_type = variable_type_mixed

        # input variables' boundaries
        if variable_type != "bool" or type(variable_type_mixed).__module__ == "numpy":
            if type(variable_boundaries).__module__ != "numpy":
                raise ValueError("variable_boundaries must be numpy array")

            if len(variable_boundaries) != self.dim:
                raise ValueError("variable_boundaries must have a length equal to dimension.")

            for i in variable_boundaries:
                assert len(i) == 2, "\n boundary for each variable must be a tuple of length two."
                assert i[0] <= i[1], "\n lower_boundaries must be smaller than upper_boundaries [lower,upper]"
            self.var_bound = variable_boundaries
        else:
            self.var_bound = np.array([[0, 1]] * self.dim)

        # input algorithm's parameters
        self.param = algorithm_parameters.__dict__

        self.pop_s = int(self.param["population_size"])

        if not (0 <= self.param["parents_portion"] <= 1):
            raise ValueError("parents_portion must be in range [0,1]")

        self.par_s = int(self.param["parents_portion"] * self.pop_s)
        trl = self.pop_s - self.par_s
        if trl % 2 != 0 and trl != 2:
            self.par_s += 1

        self.prob_mut = self.param["mutation_probability"]

        if not (0 <= self.prob_mut <= 1):
            raise ValueError("mutation_probability must be in range [0,1]")

        self.prob_cross = self.param["crossover_probability"]
        assert self.prob_cross <= 1, "mutation_probability must be less than or equal to 1"
        assert self.prob_cross >= 0, "mutation_probability must be greater than or equal to 0"

        assert self.param["elit_ratio"] <= 1, "elit_ratio must be less than or equal to 1"
        assert self.param["elit_ratio"] >= 0, "elit_ratio must be greater than or equal to 0"

        trl = self.pop_s * self.param["elit_ratio"]
        if trl < 1 and self.param["elit_ratio"] > 0:
            self.num_elit = 1
        else:
            self.num_elit = int(trl)

        assert self.par_s >= self.num_elit, "\n number of parents must be greater than number of elits"

        if self.param["max_num_iteration"] is None:
            self.iterate = 0
            for i in range(self.dim):
                if self.var_type[i] == "int":
                    self.iterate += (self.var_bound[i][1] - self.var_bound[i][0]) * self.dim * (100 / self.pop_s)
                else:
                    self.iterate += (self.var_bound[i][1] - self.var_bound[i][0]) * 50 * (100 / self.pop_s)
            self.iterate = int(self.iterate)
            max_iterations = 10000000
            if (self.iterate * self.pop_s) > max_iterations:
                self.iterate = max_iterations / self.pop_s
        else:
            self.iterate = int(self.param["max_num_iteration"])

        self.c_type = self.param["crossover_type"]
        assert self.c_type in {
            "uniform",
            "one_point",
            "two_point",
        }, "\n crossover_type must 'uniform', 'one_point', or 'two_point' Enter string"

        self.stop_mniwi = False
        if self.param["max_iteration_without_improv"] is None:
            self.mniwi = self.iterate + 1
        else:
            self.mniwi = int(self.param["max_iteration_without_improv"])

    def run(self):
        # Initial Population
        self.integers = np.where(self.var_type == "int")
        self.reals = np.where(self.var_type == "real")

        pop = np.array([np.zeros(self.dim + 1)] * self.pop_s)
        solo = np.zeros(self.dim + 1)
        var = np.zeros(self.dim)

        for p in range(self.pop_s):
            for i in self.integers[0]:
                var[i] = np.random.randint(self.var_bound[i][0], self.var_bound[i][1] + 1)
                solo[i] = var[i].copy()
            for i in self.reals[0]:
                var[i] = self.var_bound[i][0] + np.random.random() * (self.var_bound[i][1] - self.var_bound[i][0])
                solo[i] = var[i].copy()

            obj = self.sim(var)
            solo[self.dim] = obj
            pop[p] = solo.copy()

        # Report
        self.report = []
        self.test_obj = obj
        self.best_variable = var.copy()
        self.best_function = obj

        t = 1
        counter = 0
        while t <= self.iterate:
            if self.verbose is True:
                logger.info(f"Generation {t}/{self.iterate}")

            # Sort
            pop = pop[pop[:, self.dim].argsort()]

            if pop[0, self.dim] < self.best_function:
                counter = 0
                self.best_function = pop[0, self.dim].copy()
                self.best_variable = pop[0, : self.dim].copy()
                logger.info(f"Cost: {self.best_function:.4f}")
            else:
                counter += 1
            # Report
            self.report.append(pop[0, self.dim])

            # Normalizing objective function
            normobj = np.zeros(self.pop_s)

            minobj = pop[0, self.dim]
            normobj = pop[:, self.dim] + abs(minobj) if minobj < 0 else pop[:, self.dim].copy()

            maxnorm = np.amax(normobj)
            normobj = maxnorm - normobj + 1

            # Calculate probability
            sum_normobj = np.sum(normobj)
            prob = np.zeros(self.pop_s)
            prob = normobj / sum_normobj
            cumprob = np.cumsum(prob)

            # Select parents
            par = np.array([np.zeros(self.dim + 1)] * self.par_s)

            for k in range(self.num_elit):
                par[k] = pop[k].copy()
            for k in range(self.num_elit, self.par_s):
                index = np.searchsorted(cumprob, np.random.random())
                par[k] = pop[index].copy()

            ef_par_list = np.array([False] * self.par_s)
            par_count = 0
            while par_count == 0:
                for k in range(self.par_s):
                    if np.random.random() <= self.prob_cross:
                        ef_par_list[k] = True
                        par_count += 1

            ef_par = par[ef_par_list].copy()

            # New generation
            pop = np.array([np.zeros(self.dim + 1)] * self.pop_s)

            for k in range(self.par_s):
                pop[k] = par[k].copy()

            for k in range(self.par_s, self.pop_s, 2):
                r1 = np.random.randint(0, par_count)
                r2 = np.random.randint(0, par_count)
                pvar1 = ef_par[r1, : self.dim].copy()
                pvar2 = ef_par[r2, : self.dim].copy()

                ch = self.cross(pvar1, pvar2, self.c_type)
                ch1 = ch[0].copy()
                ch2 = ch[1].copy()

                ch1 = self.mut(ch1)
                ch2 = self.mutmidle(ch2, pvar1, pvar2)
                solo[: self.dim] = ch1.copy()
                obj = self.sim(ch1)
                solo[self.dim] = obj
                pop[k] = solo.copy()
                solo[: self.dim] = ch2.copy()
                obj = self.sim(ch2)
                solo[self.dim] = obj
                pop[k + 1] = solo.copy()

            t += 1
            if counter > self.mniwi:
                pop = pop[pop[:, self.dim].argsort()]
                if pop[0, self.dim] >= self.best_function:
                    t = self.iterate
                    time.sleep(2)
                    t += 1
                    self.stop_mniwi = True

        # Sort
        pop = pop[pop[:, self.dim].argsort()]

        if pop[0, self.dim] < self.best_function:
            self.best_function = pop[0, self.dim].copy()
            self.best_variable = pop[0, : self.dim].copy()
            logger.info(f"Cost: {self.best_function:.4f}")

        # Report
        self.report.append(pop[0, self.dim])

        self.output_dict = {
            "variable": self.best_variable,
            "function": self.best_function,
        }
        if self.verbose is True:
            show = " " * 100
            sys.stdout.write(f"\r{show}")
            sys.stdout.write(f"\r The best solution found:\n {self.best_variable}")
            sys.stdout.write(f"\n\n Objective function:\n {self.best_function}\n")
            sys.stdout.flush()

        if self.stop_mniwi is True:
            sys.stdout.write(
                "\nWarning: GA is terminated due to the" " maximum number of iterations without improvement was met!"
            )

    def cross(self, x, y, c_type):
        ofs1 = x.copy()
        ofs2 = y.copy()

        if c_type == "one_point":
            ran = np.random.randint(0, self.dim)
            for i in range(ran):
                ofs1[i] = y[i].copy()
                ofs2[i] = x[i].copy()

        if c_type == "two_point":
            ran1 = np.random.randint(0, self.dim)
            ran2 = np.random.randint(ran1, self.dim)

            for i in range(ran1, ran2):
                ofs1[i] = y[i].copy()
                ofs2[i] = x[i].copy()

        if c_type == "uniform":
            threshold = 0.5
            for i in range(self.dim):
                ran = np.random.random()
                if ran < threshold:
                    ofs1[i] = y[i].copy()
                    ofs2[i] = x[i].copy()

        return np.array([ofs1, ofs2])

    def mut(self, x):
        for i in self.integers[0]:
            ran = np.random.random()
            if ran < self.prob_mut:
                x[i] = np.random.randint(self.var_bound[i][0], self.var_bound[i][1] + 1)

        for i in self.reals[0]:
            ran = np.random.random()
            if ran < self.prob_mut:
                x[i] = self.var_bound[i][0] + np.random.random() * (self.var_bound[i][1] - self.var_bound[i][0])

        return x

    def mutmidle(self, x, p1, p2):
        for i in self.integers[0]:
            ran = np.random.random()
            if ran < self.prob_mut:
                if p1[i] < p2[i]:
                    x[i] = np.random.randint(p1[i], p2[i])
                elif p1[i] > p2[i]:
                    x[i] = np.random.randint(p2[i], p1[i])
                else:
                    x[i] = np.random.randint(self.var_bound[i][0], self.var_bound[i][1] + 1)

        for i in self.reals[0]:
            ran = np.random.random()
            if ran < self.prob_mut:
                if p1[i] < p2[i]:
                    x[i] = p1[i] + np.random.random() * (p2[i] - p1[i])
                elif p1[i] > p2[i]:
                    x[i] = p2[i] + np.random.random() * (p1[i] - p2[i])
                else:
                    x[i] = self.var_bound[i][0] + np.random.random() * (self.var_bound[i][1] - self.var_bound[i][0])
        return x

    def evaluate(self):
        return self.f(self.temp)

    def sim(self, X):
        self.temp = X.copy()
        return self.evaluate()



from sklearn.metrics import accuracy_score, mean_squared_error


def surrogate_accuracy_score(y_pred, y_surrogate):
    return accuracy_score(y_pred, y_surrogate)


def surrogate_mean_squared_error(y_pred, y_surrogate):
    return mean_squared_error(y_pred, y_surrogate)



"""
The :mod:`holisticai.datasets` module includes dataloaders for quick experimentation
"""

from holisticai.datasets._dataloaders import (
    load_adult,
    load_heart,
    load_last_fm,
    load_law_school,
    load_student,
    load_us_crime,
)
from holisticai.datasets._dataset import DataLoader, Dataset, DatasetDict, GroupByDataset, concatenate_datasets
from holisticai.datasets._load_dataset import load_dataset

__all__ = [
    "load_dataset",
    "Dataset",
    "DatasetDict",
    "GroupByDataset",
    "DataLoader",
    "concatenate_datasets",
    "load_adult",
    "load_last_fm",
    "load_law_school",
    "load_heart",
    "load_student",
    "load_us_crime",
]



import numpy as np


def unconventional_numeric_config():
    p_attr = np.array(["A", "A", "A", "A", "B", "B", "B", "B", "C", "C"])
    y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0]) + 1
    y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1]) + 1
    return p_attr, y_pred, y_true


def numeric_config():
    p_attr = np.array(["A", "A", "A", "A", "B", "B", "B", "B", "C", "C"])
    y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])  # , dtype=np.float32)
    y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])  # , dtype=np.float32)
    return p_attr, y_pred, y_true


def numeric_nan_config(y_pred_nan=True):
    p_attr = np.array(["A", "A", "A", "A", "B", "B", "B", "B", "C", "C"])
    y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    if y_pred_nan:
        y_pred = np.array([0, 1, 2, 0, 1, 2, np.nan, 1, 2, 0])  # , dtype=np.float32)
    else:
        y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, np.nan, 1])
    return p_attr, y_pred, y_true


def str_config():
    p_attr = np.array(["A", "A", "A", "A", "B", "B", "B", "B", "C", "C"])
    classes = ["a", "b", "c"]
    i2c = dict(zip(range(len(classes)), classes))
    y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    y_pred = np.array([i2c[y] for y in y_pred])
    y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    y_true = np.array([i2c[y] for y in y_true])
    return p_attr, y_pred, y_true



import numpy as np
import pandas as pd
import scipy.optimize as opt
from holisticai.bias.mitigation.inprocessing.commons._conventions import PRECISION
from sklearn import clone


class Lagrangian:
    """Operations related to the Lagrangian"""

    def __init__(
        self, X: np.ndarray, y: np.ndarray, estimator, constraints, B: float, opt_lambda: bool = True, **kwargs
    ):
        self.constraints = constraints
        self.constraints.load_data(X, y, **kwargs)
        self.obj = self.constraints.default_objective()
        self.obj.load_data(X, y, **kwargs)
        self.estimator = estimator
        self.B = B
        self.opt_lambda = opt_lambda
        self.hs = pd.Series(dtype="object")
        self.predictors = pd.Series(dtype="object")
        self.errors = pd.Series(dtype="float64")
        self.gammas = pd.DataFrame()
        self.lambdas = pd.DataFrame()
        self.last_linprog_n_hs = 0
        self.last_linprog_result = None

    def _eval(self, Q, lambda_vec):
        if callable(Q):
            error = self.obj.gamma(Q)[0]
            gamma = self.constraints.gamma(Q)
        else:
            error = self.errors[Q.index].dot(Q)
            gamma = self.gammas[Q.index].dot(Q)

        lambda_projected = self.constraints.project_lambda(lambda_vec) if self.opt_lambda else lambda_vec
        constraint_violation = gamma - self.constraints.bound()
        L = error + np.sum(lambda_projected * constraint_violation)
        max_constraint = np.max(constraint_violation)
        L_high = error if max_constraint <= 0 else error + self.B * max_constraint
        return L, L_high, gamma, error

    def eval_gap(self, Q, lambda_hat, nu):
        L, L_high, gamma, error = self._eval(Q, lambda_hat)
        result = _GapResult(L, L, L_high, gamma, error)
        for mul in [1.0, 2.0, 5.0, 10.0]:
            h_hat, h_hat_idx = self.best_h(mul * lambda_hat)
            L_low_mul, _, _, _ = self._eval(pd.Series({h_hat_idx: 1.0}), lambda_hat)
            if L_low_mul < result.L_low:
                result.L_low = L_low_mul
            if result.gap() > nu + PRECISION:
                break
        return result

    def solve_linprog(self, nu):
        n_hs = len(self.hs)
        if self.last_linprog_n_hs == n_hs:
            return self.last_linprog_result

        linprog_data = self._prepare_linprog_data(n_hs, len(self.constraints.index))
        result = opt.linprog(**linprog_data, method="highs")
        Q = pd.Series(result.x[:-1], self.hs.index)

        dual_linprog_data = self._prepare_dual_linprog_data(linprog_data)
        result_dual = opt.linprog(**dual_linprog_data, method="highs")
        lambda_vec = pd.Series(result_dual.x[:-1], self.constraints.index)

        self.last_linprog_n_hs = n_hs
        self.last_linprog_result = (Q, lambda_vec, self.eval_gap(Q, lambda_vec, nu))
        return self.last_linprog_result

    def _prepare_linprog_data(self, n_hs, n_constraints):
        c = np.append(self.errors.values, self.B)
        A_ub = np.hstack((self.gammas.sub(self.constraints.bound(), axis=0).values, -np.ones((n_constraints, 1))))
        b_ub = np.zeros(n_constraints)
        A_eq = np.hstack((np.ones((1, n_hs)), np.zeros((1, 1))))
        b_eq = np.ones(1)
        return {"c": c, "A_ub": A_ub, "b_ub": b_ub, "A_eq": A_eq, "b_eq": b_eq}

    def _prepare_dual_linprog_data(self, linprog_data):
        n_constraints = len(linprog_data["b_ub"])
        dual_c = np.append(linprog_data["b_ub"], -linprog_data["b_eq"])
        dual_A_ub = np.hstack((-linprog_data["A_ub"].T, linprog_data["A_eq"].T))
        dual_b_ub = linprog_data["c"]
        dual_bounds = [(None, None) if i == n_constraints else (0, None) for i in range(n_constraints + 1)]
        return {"c": dual_c, "A_ub": dual_A_ub, "b_ub": dual_b_ub, "bounds": dual_bounds}

    def _call_oracle(self, lambda_vec):
        signed_weights = self.obj.signed_weights() + self.constraints.signed_weights(lambda_vec)
        y = (
            (signed_weights > 0).astype(int)
            if self.constraints.PROBLEM_TYPE == "classification"
            else self.constraints.y_as_series
        )
        w = np.abs(signed_weights)
        sample_weight = self.constraints.total_samples * w / np.sum(w)
        estimator = clone(self.estimator, safe=False)
        estimator.fit(self.constraints.X, y, sample_weight=sample_weight)
        return estimator

    def best_h(self, lambda_vec):
        classifier = self._call_oracle(lambda_vec)

        def h(X):
            pred = classifier.predict(X)
            return pred.flatten() if hasattr(pred, "flatten") else pred

        h_error = self.obj.gamma(h).iloc[0]
        h_gamma = self.constraints.gamma(h)
        h_value = h_error + np.dot(h_gamma, lambda_vec)

        if not self.hs.empty:
            values = self.errors + np.dot(self.gammas.T, lambda_vec)
            best_idx = values.idxmin()
            best_value = values[best_idx]
        else:
            best_idx = -1
            best_value = np.inf

        if h_value < best_value - PRECISION:
            h_idx = len(self.hs)
            self.hs.at[h_idx] = h
            self.predictors.at[h_idx] = classifier
            self.errors.at[h_idx] = h_error
            self.gammas[h_idx] = h_gamma
            self.lambdas[h_idx] = lambda_vec.copy()
            best_idx = h_idx

        return self.hs[best_idx], best_idx


class _GapResult:
    """The result of a duality gap computation."""

    def __init__(self, L, L_low, L_high, gamma, error):
        self.L = L
        self.L_low = L_low
        self.L_high = L_high
        self.gamma = gamma
        self.error = error

    def gap(self):
        return max(self.L - self.L_low, self.L_high - self.L)



import pandas as pd


def check_valid_output_type(y_true: pd.Series):
    if y_true.dtype.kind in ["i", "u", "O"]:
        return "classification"
    if y_true.dtype.kind in ["f"]:
        return "regression"
    raise ValueError(f"The target variable must be either continuous or categorical, found: {y_true.dtype}")



from typing import Literal

import numpy as np
from holisticai.bias.mitigation.postprocessing.lp_debiaser.multiclass_balancer.algorithm import (
    MulticlassBalancerAlgorithm,
)
from holisticai.utils.transformers.bias import BMPostprocessing as BMPost
from holisticai.utils.transformers.bias import SensitiveGroups


class LPDebiaserMulticlass(BMPost):
    """
    Linear Programming Debiaser [1]_ is a postprocessing algorithms designed to debias pretrained classifiers.\
    The algorithm use constraints such as Equalized Odds and Equalized Opportunity.\
    This technique extends LPDebiaserBinary for multiclass classification.

    Parameters
    ----------
    constraint : str
        Strategy used to evalute the cost function  The available contraints  are:\
        "EqualizedOdds", "EqualizedOpportunity"

    loss : str
        The loss function to optimize: "macro", "micro". The default is "macro"

    References
    ----------
        .. [1] Putzel, Preston, and Scott Lee. "Blackbox Post-Processing for Multiclass Fairness."\
        arXiv preprint arXiv:2201.04461 (2022).
    """

    CONSTRAINT = Literal["EqualizedOdds", "EqualizedOpportunity"]
    OBJ_FUN = Literal["macro", "micro"]

    def __init__(
        self,
        constraint: CONSTRAINT = "EqualizedOdds",
        loss: OBJ_FUN = "macro",
    ):
        self.constraint = constraint
        self.loss = loss
        self._sensgroups = SensitiveGroups()

    def fit(
        self,
        y: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Compute parameters for Linear Programming Debiaser.\
        For binary classification y_pred or y_proba can be used.\
        For Multiclass classification only y_pred must be used.\

        Parameters
        ----------
        y : array-like
            Target vector
        y_pred : array-like
            Predicted label vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        Self
        """
        params = self._load_data(y=y, y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y = params["y"]
        y_pred = params["y_pred"]

        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self._sensgroups.fit_transform(sensitive_features, convert_numeric=True)

        constraints_catalog, objective_catalog = self._get_catalogs()

        constraint = constraints_catalog[self.constraint]()

        objective = objective_catalog[self.loss]()

        self.algorithm = MulticlassBalancerAlgorithm(constraint=constraint, objective=objective)

        self.algorithm.fit(y_true=y, y_pred=y_pred, p_attr=p_attr)
        return self

    def transform(
        self,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Apply transform function to predictions and likelihoods

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examples,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dict
            A dictionnary with new predictions
        """

        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]

        sensitive_features = np.stack([group_a, group_b], axis=1)
        p_attr = self._sensgroups.transform(sensitive_features, convert_numeric=True)
        new_y_pred = self.algorithm.predict(y_pred=y_pred, p_attr=p_attr)
        return {"y_pred": new_y_pred}

    def fit_transform(
        self,
        y: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit and transform

        Parameters
        ----------
        y : array-like
            Target vector
        y_pred : array-like
            Predicted vector (nb_examples,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        -------
        dict
            A dictionnary with new predictions
        """
        return self.fit(
            y=y,
            y_pred=y_pred,
            group_a=group_a,
            group_b=group_b,
        ).transform(y_pred=y_pred, group_a=group_a, group_b=group_b)

    def _get_catalogs(self):
        from holisticai.bias.mitigation.postprocessing.lp_debiaser.multiclass_balancer.constraints import (
            EqualizedOdds,
            EqualizedOpportunity,
            MacroLosses,
            MicroLosses,
            Strict,
        )

        cons_cat = {
            "EqualizedOdds": EqualizedOdds,
            "EqualizedOpportunity": EqualizedOpportunity,
            "Strict": Strict,
        }

        obj_cat = {"macro": MacroLosses, "micro": MicroLosses}

        return cons_cat, obj_cat



import jax.numpy as jnp
from holisticai.utils.transformers.bias import BMPreprocessing as BMPre
from jax import jit


@jit
def _fit(X, sensitive_features, sensitive_mean):
    sensitive_features_center = sensitive_features - sensitive_mean
    beta = jnp.linalg.lstsq(sensitive_features_center, X, rcond=None)[0]
    return beta


@jit
def _transform(X, sensitive_features, beta, alpha, sensitive_mean):
    sensitive_features_center = sensitive_features - sensitive_mean
    x_filtered = X - jnp.dot(sensitive_features_center, beta)
    return alpha * x_filtered + (1 - alpha) * X


class CorrelationRemover(BMPre):
    """
    CorrelationRemover applies a linear transformation to the non-sensitive feature columns\
    in order to remove their correlation with the sensitive feature columns while retaining\
    as much information as possible (as measured by the least-squares error).

    Parameters
    ----------
    alpha : float, optional
        parameter to control how much to filter, for alpha=1.0 we filter out
        all information while for alpha=0.0 we don't apply any. Default is 1.

    Notes
    -----
    This method will change the original dataset by removing all correlation with sensitive\
    values. Note that the lack of correlation does not imply anything about statistical dependence.\
    Therefore, it is expected this to be most appropriate as a preprocessing step for\
    (generalized) linear models.
    """

    def __init__(self, alpha=1):
        self.alpha = alpha

    def fit(self, X: jnp.ndarray, group_a: jnp.ndarray, group_b: jnp.ndarray):
        """
        Learn the projection required to make the dataset uncorrelated with groups (group_a and group_b).

        Parameters
        ----------
        X : matrix-like
            Input data
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        ------
        Self
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        x = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        sensitive_features = jnp.stack([group_a, group_b], axis=1).astype(jnp.int32)
        self.sensitive_mean_ = jnp.mean(sensitive_features, axis=0)
        self.beta_ = _fit(x, sensitive_features, self.sensitive_mean_)
        self.x_shape_ = x.shape
        return self

    def transform(self, X: jnp.ndarray, group_a: jnp.ndarray, group_b: jnp.ndarray):
        """
        Transform X by applying the correlation remover.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
            np.ndarray
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        x = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        sensitive_features = jnp.stack([group_a, group_b], axis=1).astype(jnp.int32)
        x_filtered = _transform(x, sensitive_features, self.beta_, self.alpha, self.sensitive_mean_)
        return x_filtered

    def fit_transform(self, X: jnp.ndarray, group_a: jnp.ndarray, group_b: jnp.ndarray):
        """
        Fit and transform

        Parameters
        ----------
        X : matrix-like
            Input data
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        ------
            Self
        """
        self.fit(X, group_a, group_b)
        return self.transform(X, group_a, group_b)



from __future__ import annotations

import logging
from typing import Literal, get_args

import numpy as np
import pandas as pd
from holisticai.utils import Importances, ModelProxy
from holisticai.utils.feature_importances import compute_permutation_feature_importance
from sklearn.feature_selection import (
    SelectPercentile,
    VarianceThreshold,
    f_classif,
    f_regression,
)

logger = logging.getLogger(__name__)
SelectorbyData = Literal["Percentile", "Variance"]
SelectorbyImportance = Literal["FImportance"]
SelectorType = Literal[SelectorbyData, SelectorbyImportance]


def get_score_fn(learning_task: str):
    if learning_task in ["binary_classification", "multi_classification"]:
        return f_classif
    if learning_task == "regression":
        return f_regression
    raise ValueError(f"Learning task {learning_task} not supported")


class SelectorsHandler:
    def __init__(self, proxy: ModelProxy, selector_types: list[SelectorType]):
        self.proxy = proxy
        self.percentiles = [80, 90]
        self.variance_thresholds = [80, 90]
        self.importance_thresholds = [80, 90]
        self.selector_types = selector_types
        self.selectors = {}

    def _get_selectors_from_data(self, selector_type: SelectorbyData):
        percent_range = np.array(self.percentiles)
        variance_thresholds = np.array(self.variance_thresholds)

        if selector_type == "Percentile":
            methods = {}
            for p in percent_range:
                score_fn = get_score_fn(self.proxy.learning_task)
                methods[f"Percentile >{p}"] = SelectPercentile(score_fn, percentile=p)
            return methods

        if selector_type == "Variance":
            methods = {}
            for p in variance_thresholds:
                methods[f"Variance >{p}"] = VarianceThreshold(threshold=p / 100)
            return methods
        raise ValueError(f"Selector type {selector_type} not supported")

    def _get_selectors_from_importances(self, selector_type: SelectorbyImportance):
        if selector_type == "FImportance":
            methods = {}
            for p in self.importance_thresholds:
                methods[f"FImportance >{p}"] = SelectFromFeatureImportance(threshold=p / 100)
            return methods
        raise ValueError(f"Selector type {selector_type} not supported")

    def fit(self, X: pd.DataFrame, y: pd.Series):
        selectors_by_data = {}
        selectors_by_importance = {}

        importances = None
        for selector_type in self.selector_types:
            if selector_type in get_args(SelectorbyData):
                selectors_by_data.update(self._get_selectors_from_data(selector_type))

            if selector_type in get_args(SelectorbyImportance):
                selectors_by_importance.update(self._get_selectors_from_importances(selector_type))

        for sn, selector in selectors_by_data.items():
            logger.info(f"Fitting selector {sn}")
            Xt = (
                X.select_dtypes(include=["number"])
                if any(sn.startswith(name) for name in ["Percentile", "Variance"])
                else X
            )
            selector.fit(Xt, y)

        for sn, selector in selectors_by_importance.items():
            logger.info(f"Fitting selector {sn}")
            if importances is None:
                importances = compute_permutation_feature_importance(proxy=self.proxy, X=X, y=y)
            selector.fit(importances)

        self.selectors = {**selectors_by_data, **selectors_by_importance}
        return self

    def __iter__(self):
        return iter(self.selectors.items())

    def get_feature_indexes(self):
        return {f: s.get_feature_names_out() for f, s in self.selectors.items()}


class SelectFromFeatureImportance:
    def __init__(self, threshold: float):
        """
        Parameters
        ----------
        feat_imp : FeatureImportance
            An FeatureImportance instance.

        threshold : float
            Used to select the minimum amount of features that cover this importance fraction.
        """
        self.threshold = threshold

    def fit(self, importances: Importances):
        """
        Create a list of top features based on the threshold.
        Return
        ------
            Self
        """
        self._importances = importances
        top_importances = self._importances.top_alpha(self.threshold)
        self.top_feature_names = top_importances.feature_names
        self.top_feature_index = [
            int(i) for i, fi in enumerate(self._importances.feature_names) if fi in self.top_feature_names
        ]
        return self

    def transform(self, X: pd.DataFrame):
        """
        Parameters
        ----------
        x: matrix-like
            input data (n_examples, n_features)

        Return
        ------
            matrix-like
            x matrix with only the selected features.
        """
        return X[self.top_feature_names]

    def get_feature_names_out(self):
        "Return list of columns index"
        return self.top_feature_index

    def fit_transform(self, X: pd.DataFrame, y: pd.Series):
        return self.fit(X, y).transform(X)



from __future__ import annotations

from typing import TYPE_CHECKING

from holisticai.bias.mitigation.postprocessing.mcmf_clustering.algorithm import Algorithm
from holisticai.utils.transformers.bias import BMPostprocessing as BMPost

if TYPE_CHECKING:
    import numpy as np


class MCMF(BMPost):
    """
    Minimal Cluster Modification for Fairnes (MCMF) [1]_ is focused on the minimal change it so that the clustering is still\
    of good quality and fairer.

    Parameters
    ----------
    metric : str
        Measure function used in the objective function.
        The metrics available are:
        ["constant", "L1", "L2"]

    solver : str
        Algorithm name used to solve the standard form problem. Solver supported must depend of your scipy package version.
        for scipy 1.9.0 the solvers available are:
        ["highs", "highs-ds", "highs-ipm"]

    group_mode : str
        Set what groups will be fitted: ['a', 'b', 'ab']
    verbose : int
        If > 0 , then print logs.

    References
    ----------
        .. [1] Davidson, Ian, and S. S. Ravi. "Making existing clusterings fairer: Algorithms, complexity results and insights."\
        Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 34. No. 04. 2020.
    """

    def __init__(self, metric: str = "L1", solver: str = "highs", group_mode="a", verbose=0):
        self.metric = metric
        self.group_mode = group_mode
        self.solver = solver
        self.verbose = verbose
        self.algorithm = Algorithm(metric=metric, solver=solver, verbose=verbose)

    def fit(self):
        """
        Fit the MCMF algorithm.

        Returns
        -------
        self
        """
        return self

    def fit_transform(
        self,
        X: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        centroids: np.ndarray,
    ) -> dict[str, np.ndarray]:
        """
        Fit and transform the MCMF algorithm.

        Description
        ----------
        Fit the MCMF algorithm and transform the predicted vector.

        Parameters
        ----------
        X : matrix-like
            Input matrix (nb_examples, nb_features)
        y_pred : array-like
            Predicted vector (nb_examples,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        centroids : ndarray
            Centroids array

        Returns
        -------
        dict
            A dictionary with new predictions
        """
        return self.transform(X, y_pred, group_a, group_b, centroids)

    def transform(
        self,
        X: np.ndarray,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        centroids: np.ndarray,
    ) -> dict[str, np.ndarray]:
        """
        Modify y_pred using MCMF algorithm.

        Description
        ----------
        Build parameters for the objetive function and call the solver to find the algorithm parameters.

        Parameters
        ----------
        X : matrix-like
            Input matrix (nb_examples, nb_features)
        y_pred : array-like
            Predicted vector (nb_examples,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        centroids : ndarray
            Centroids array

        Returns
        -------
        dict
            A dictionary with new predictions
        """
        params = self._load_data(X=X, y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        X = params["X"]
        y_pred = params["y_pred"]

        if isinstance(centroids, str):
            centroids = getattr(self.estimator_hdl.estimator, centroids)

        if self.group_mode in ["a", "b"]:
            group = group_a if self.group_mode == "a" else group_b
            new_y_pred = self.algorithm.transform(X=X, y_pred=y_pred, group=group, centroids=centroids)
        elif self.group_mode == "ab":
            new_y_pred = y_pred.copy()
            for group in [group_a, group_b]:
                new_y_pred = self.algorithm.transform(X=X, y_pred=new_y_pred, group=group, centroids=centroids)

        return {"y_pred": new_y_pred}



"""
The :mod:`holisticai.bias.mitigation` module includes preprocessing, inprocessing and postprocessing bias mitigation algorithms.
"""

# inprocessing algorithm classes
from holisticai.bias.mitigation.inprocessing import (
    AdversarialDebiasing,
    BlindSpotAwareMF,
    DebiasingLearningMF,
    ExponentiatedGradientReduction,
    FairKCenterClustering,
    FairKMedianClustering,
    FairletClustering,
    FairRec,
    GridSearchReduction,
    MetaFairClassifier,
    PopularityPropensityMF,
    PrejudiceRemover,
    VariationalFairClustering,
)

# postprocessing algorithm classes
from holisticai.bias.mitigation.postprocessing import (
    MCMF,
    CalibratedEqualizedOdds,
    DebiasingExposure,
    EqualizedOdds,
    FairTopK,
    LPDebiaserBinary,
    LPDebiaserMulticlass,
    MLDebiaser,
    PluginEstimationAndCalibration,
    RejectOptionClassification,
    WassersteinBarycenter,
)

# preprocessing algorithm classes
from holisticai.bias.mitigation.preprocessing import (
    CorrelationRemover,
    FairletClusteringPreprocessing,
    LearningFairRepresentation,
    Reweighing,
)

# all
__all__ = [
    "AdversarialDebiasing",
    "CorrelationRemover",
    "Reweighing",
    "LearningFairRepresentation",
    "ExponentiatedGradientReduction",
    "GridSearchReduction",
    "CalibratedEqualizedOdds",
    "EqualizedOdds",
    "RejectOptionClassification",
    "WassersteinBarycenter",
    "MLDebiaser",
    "LPDebiaserBinary",
    "LPDebiaserMulticlass",
    "PluginEstimationAndCalibration",
    "PrejudiceRemover",
    "MetaFairClassifier",
    "VariationalFairClustering",
    "FairletClustering",
    "FairletClusteringPreprocessing",
    "FairKCenterClustering",
    "FairKMedianClustering",
    "BlindSpotAwareMF",
    "DebiasingLearningMF",
    "PopularityPropensityMF",
    "FairRec",
    "DebiasingExposure",
    "FairTopK",
    "MCMF",
]

import importlib
from typing import Literal

networkx_spec = importlib.util.find_spec("networkx")
if networkx_spec is not None:
    from holisticai.bias.mitigation.postprocessing import DisparateImpactRemoverRS
    from holisticai.bias.mitigation.preprocessing import DisparateImpactRemover

__all__ += ["DisparateImpactRemoverRS", "DisparateImpactRemover"]

cvxpy_spec = importlib.util.find_spec("cvxpy")
if cvxpy_spec is not None:
    from holisticai.bias.mitigation.inprocessing.fair_scoring_classifier.transformer import FairScoreClassifier
__all__ += ["FairScoreClassifier"]

MITIGATOR_NAME = Literal[
    "CorrelationRemover",
    "Reweighing",
    "LearningFairRepresentation",
    "ExponentiatedGradientReduction",
    "GridSearchReduction",
    "CalibratedEqualizedOdds",
    "EqualizedOdds",
    "RejectOptionClassification",
    "WassersteinBarycenter",
    "MLDebiaser",
    "LPDebiaserBinary",
    "LPDebiaserMulticlass",
    "PluginEstimationAndCalibration",
    "PrejudiceRemover",
    "MetaFairClassifier",
    "VariationalFairClustering",
    "FairletClustering",
    "FairletClusteringPreprocessing",
    "FairKCenterClustering",
    "FairKMedianClustering",
    "BlindSpotAwareMF",
    "DebiasingLearningMF",
    "PopularityPropensityMF",
    "FairRec",
    "FairScoreClassifier",
    "DebiasingExposure",
    "FairTopK",
    "MCMF",
    "AdversarialDebiasing",
    "DisparateImpactRemoverRS",
    "DisparateImpactRemover",
]
